/*
 *  Butterfly.h
 * 
 *  Author: Adapted from code for Loop subdivision (by Computer Graphics Group, RWTH Aachen)
 *
 *  Content: Butterfly subdivision.
 *
 */


//=============================================================================
//
//  CLASS ButterflyT
//
//=============================================================================

#ifndef BUTTERFLYT_H
#define BUTTERFLYT_H


//== INCLUDES =================================================================

#include <OpenMesh/Core/System/config.hh>
#include <OpenMesh/Tools/Subdivider/Uniform/SubdividerT.hh>
#include <OpenMesh/Core/Utils/vector_cast.hh>
// -------------------- STL
#include <vector>
#if defined(OM_CC_MIPS)
#  include <math.h>
#else
#  include <cmath>
#endif


//== CLASS DEFINITION =========================================================

/** Butterfly subdivision algorithm.
 */
template <typename MeshType, typename RealType = float>
class ButterflyT : public OpenMesh::Subdivider::Uniform::SubdividerT<MeshType, RealType>
{
public:

  typedef RealType                                real_t;
  typedef MeshType                                mesh_t;
  typedef OpenMesh::Subdivider::Uniform::SubdividerT< mesh_t, real_t >  parent_t;

  typedef std::vector< std::vector<real_t> >        weights_t;

  // butterfly algorithm
  enum ButterflyMode
  {
      eOriginal,        // original butterfly interpolation scheme
      eModified,        // modified butterfly interpolation scheme
  };

public:


  ButterflyT(ButterflyMode mode = eOriginal) : parent_t(), max_valence_(16)
  {
      mode_ = mode;
      init_weights();
  }


  ButterflyT( mesh_t& _m, ButterflyMode mode = eOriginal) : parent_t(_m), , max_valence_(16)
  {
      mode_ = mode;
      init_weights();
  }

  ~ButterflyT()
  {
      // a hack, but somehow base class destructor fails to do it
      detach();
  }


public:


  const char *name() const { return "Uniform Butterfly"; }


  // Pre-compute weights
  void init_weights(void)
  {
      weights_.resize(max_valence_-2);

      // valence 3
      weights_[0].resize(3);
      (weights_[0])[0] = real_t(5.0/12.0);
      (weights_[0])[1] = real_t(-1.0/12.0);
      (weights_[0])[2] = real_t(-1.0/12.0);

      // valence 4
      weights_[1].resize(4);
      (weights_[1])[0] = real_t(3.0/8.0);
      (weights_[1])[1] = real_t(0.0);
      (weights_[1])[2] = real_t(-1.0/8.0);
      (weights_[1])[3] = real_t(0.0);

      for(int i=5; i <= max_valence_; i++)
      {
          weights_[i-3].resize(i);
          compute_weight(weights_[i-3], i);
      }
  }


protected:

  
  bool prepare( mesh_t& _m )
  {
    _m.add_property( ep_pos_ );
    return true;
  }
  

  bool cleanup( mesh_t& _m )
  {
    _m.remove_property( ep_pos_ );
    return true;
  }


  void subdivide_original( mesh_t& _m, size_t _n)
  {
      typename mesh_t::FaceIter   fit, f_end;
      typename mesh_t::EdgeIter   eit, e_end;
      
      // Do _n subdivisions
      for (size_t i=0; i < _n; ++i)
      {
          // Compute position for new vertices and store them in the edge property
          for (eit=_m.edges_begin(); eit != _m.edges_end(); ++eit)
              compute_midpoint( _m, eit.handle() );
          
          // Attention! Creating new edges, hence make sure the loop ends correctly.
          e_end = _m.edges_end();
          for (eit=_m.edges_begin(); eit != e_end; ++eit)
              split_edge(_m, eit.handle() );
          
          // Commit changes in topology and reconsitute consistency
          
          // Attention! Creating new faces, hence make sure the loop ends correctly.
          f_end   = _m.faces_end();
          for (fit = _m.faces_begin(); fit != f_end; ++fit)
              split_face(_m, fit.handle() );
      
#if defined(_DEBUG) || defined(DEBUG)
      // Now we have an consistent mesh!
      assert( OpenMesh::Utils::MeshCheckerT<mesh_t>(_m).check() );
#endif
    }
  }

  void subdivide_modified( mesh_t& _m, size_t _n)
  {
      typename mesh_t::FaceIter   fit, f_end;
      typename mesh_t::EdgeIter   eit, e_end;
      typename mesh_t::VertexIter vit;
      typename mesh_t::HalfedgeHandle hh;

      OpenMesh::VPropHandleT< int > vp_val;
      int val1, val2;
      typename mesh_t::Point pt;

      // vertex valence counter
      _m.add_property(vp_val);
      
      // Do _n subdivisions
      for (size_t i=0; i < _n; ++i)
      {
          compute_valence(_m, vp_val);

          // Compute position for new vertices and store them in the edge property
          for (eit=_m.edges_begin(); eit != _m.edges_end(); ++eit)
          {
              // check if vertices of edge are extraordinary points (valence != 6)
              hh = _m.halfedge_handle(eit, 0);
              val1 = _m.property(vp_val, _m.to_vertex_handle(hh));
              val2 = _m.property(vp_val, _m.from_vertex_handle(hh));

              if ((val1 == 6) && (val2 == 6))
              {
                  // both vertices are ordinary
                  compute_midpoint( _m, eit.handle() );
              }
              else
              {
                  pt.vectorize(real_t(0.0));

                  // one or both vertices are extraordinary points
                  if (val2 != 6)
                  {
                      pt = pt + extra_midpoint(_m, hh, val2);
                  }

                  if (val1 != 6)
                  {
                      hh = _m.opposite_halfedge_handle(hh);
                      pt = pt + extra_midpoint(_m, hh, val1);
                  }

                  if ((val1 != 6) && (val2 != 6))
                      pt = pt * (real_t(0.5));

                  _m.property(ep_pos_, eit) = pt;
              }
          }
          
          // Attention! Creating new edges, hence make sure the loop ends correctly.
          e_end = _m.edges_end();
          for (eit=_m.edges_begin(); eit != e_end; ++eit)
              split_edge(_m, eit.handle() );
          
          // Commit changes in topology and reconsitute consistency
          
          // Attention! Creating new faces, hence make sure the loop ends correctly.
          f_end   = _m.faces_end();
          for (fit = _m.faces_begin(); fit != f_end; ++fit)
              split_face(_m, fit.handle() );
      
#if defined(_DEBUG) || defined(DEBUG)
      // Now we have an consistent mesh!
      assert( OpenMesh::Utils::MeshCheckerT<mesh_t>(_m).check() );
#endif
      }
      
      _m.remove_property(vp_val);
  }

  bool subdivide( mesh_t& _m, size_t _n)
  {
      if (mode_ == eOriginal)
          subdivide_original(_m, _n);
      else if (mode_ == eModified)
          subdivide_modified(_m, _n);

      return true;
  }

private:

  // Helper functor to compute weights for Butterfly-subdivision
  void compute_weight(std::vector<real_t> &vec, int valence)
  {
      //         1     1
      // s(i) =  - * ( - + cos(2*Pi*i/k) + 0.5*cos(4*Pi*i/k) )
      //         k     4

      for(int i=0; i < valence; i++)
      {
          double weight = (0.25 + cos(2.0*M_PI*i / valence) + (0.5*cos(4.0*M_PI*i / valence)));
          vec[i] = real_t(weight / valence);
      }
  };

  // helper function to compute valence of vertices
  void compute_valence(mesh_t& _m, OpenMesh::VPropHandleT< int >& vp_val)
  {
      typename mesh_t::VertexIter vIt, vEnd = _m.vertices_end();
      typename mesh_t::ConstVertexVertexIter cvvIt;     // vertex neighborhood iterator
      
      for(vIt = _m.vertices_begin(); vIt != vEnd; ++vIt)
      {
          int valence = 0;
          for(cvvIt = _m.cvv_iter(vIt); cvvIt; ++cvvIt)
          {
			  valence++;
		  }

          _m.property(vp_val, vIt) = valence;
      }
  }

private: // topological modifiers

  void split_face(mesh_t& _m, const typename mesh_t::FaceHandle& _fh)
  {
    typename mesh_t::HalfedgeHandle 
      heh1(_m.halfedge_handle(_fh)),
      heh2(_m.next_halfedge_handle(_m.next_halfedge_handle(heh1))),
      heh3(_m.next_halfedge_handle(_m.next_halfedge_handle(heh2)));

    // Cutting off every corner of the 6_gon
    corner_cutting( _m, heh1 );
    corner_cutting( _m, heh2 );
    corner_cutting( _m, heh3 );
  }


  void corner_cutting(mesh_t& _m, const typename mesh_t::HalfedgeHandle& _he)
  {
    // Define Halfedge Handles
    typename mesh_t::HalfedgeHandle 
      heh1(_he),
      heh5(heh1),
      heh6(_m.next_halfedge_handle(heh1));

    // Cycle around the polygon to find correct Halfedge
    for (; _m.next_halfedge_handle(_m.next_halfedge_handle(heh5)) != heh1; 
         heh5 = _m.next_halfedge_handle(heh5))
    {}

    typename mesh_t::VertexHandle
      vh1 = _m.to_vertex_handle(heh1),
      vh2 = _m.to_vertex_handle(heh5);

    typename mesh_t::HalfedgeHandle 
      heh2(_m.next_halfedge_handle(heh5)),
      heh3(_m.new_edge( vh1, vh2)),
      heh4(_m.opposite_halfedge_handle(heh3));

    /* Intermediate result
     *
     *            *
     *         5 /|\
     *          /_  \
     *    vh2> *     *
     *        /|\3   |\
     *       /_  \|4   \
     *      *----\*----\*
     *          1 ^   6
     *            vh1 (adjust_outgoing halfedge!)
     */

    // Old and new Face
    typename mesh_t::FaceHandle     fh_old(_m.face_handle(heh6));
    typename mesh_t::FaceHandle     fh_new(_m.new_face());


    // Re-Set Handles around old Face
    _m.set_next_halfedge_handle(heh4, heh6);
    _m.set_next_halfedge_handle(heh5, heh4);

    _m.set_face_handle(heh4, fh_old);
    _m.set_face_handle(heh5, fh_old);
    _m.set_face_handle(heh6, fh_old);
    _m.set_halfedge_handle(fh_old, heh4);

    // Re-Set Handles around new Face
    _m.set_next_halfedge_handle(heh1, heh3);
    _m.set_next_halfedge_handle(heh3, heh2);

    _m.set_face_handle(heh1, fh_new);
    _m.set_face_handle(heh2, fh_new);
    _m.set_face_handle(heh3, fh_new);

    _m.set_halfedge_handle(fh_new, heh1);
  }


  void split_edge(mesh_t& _m, const typename mesh_t::EdgeHandle& _eh)
  {
    typename mesh_t::HalfedgeHandle 
      heh     = _m.halfedge_handle(_eh, 0),
      opp_heh = _m.halfedge_handle(_eh, 1);

    typename mesh_t::HalfedgeHandle new_heh, opp_new_heh, t_heh;
    typename mesh_t::VertexHandle   vh;
    typename mesh_t::VertexHandle   vh1(_m.to_vertex_handle(heh));
    typename mesh_t::Point          zero(0,0,0);

    // new vertex
    vh                = _m.new_vertex( zero );

    // set position of new vertex
    _m.set_point(vh, _m.property( ep_pos_, _eh ));

    
    // Re-link mesh entities
    if (_m.is_boundary(_eh)) 
    {
      for (t_heh = heh; 
           _m.next_halfedge_handle(t_heh) != opp_heh; 
           t_heh = _m.opposite_halfedge_handle(_m.next_halfedge_handle(t_heh))) 
      {}
    } 
    else 
    {
      for (t_heh = _m.next_halfedge_handle(opp_heh); 
           _m.next_halfedge_handle(t_heh) != opp_heh; 
           t_heh = _m.next_halfedge_handle(t_heh) ) 
      {}
    }

    new_heh     = _m.new_edge(vh, vh1);
    opp_new_heh = _m.opposite_halfedge_handle(new_heh);

    _m.set_next_halfedge_handle(t_heh, opp_new_heh);
    _m.set_next_halfedge_handle(new_heh, _m.next_halfedge_handle(heh));
    _m.set_next_halfedge_handle(heh, new_heh);
    _m.set_next_halfedge_handle(opp_new_heh, opp_heh);

    if (_m.face_handle(opp_heh).is_valid()) 
    {
      _m.set_face_handle(opp_new_heh, _m.face_handle(opp_heh));
      _m.set_halfedge_handle(_m.face_handle(opp_new_heh), opp_new_heh);
    }

    _m.set_vertex_handle( heh, vh ); 
    _m.set_face_handle( new_heh, _m.face_handle(heh) );
    _m.set_halfedge_handle( vh, new_heh);
    _m.set_halfedge_handle( _m.face_handle(heh), heh );
    _m.set_halfedge_handle( vh1, opp_new_heh );

    // Never forget this, when playing with the topology
    _m.adjust_outgoing_halfedge( vh );
    _m.adjust_outgoing_halfedge( vh1 );
  }

private: // geometry helper

/*======================================================================
  = reflect point (on axis given by half_edge handle)                  =
  ======================================================================*/
  void reflect_point(mesh_t& _m, typename mesh_t::Point& _pt, typename mesh_t::HalfedgeHandle _hh)
  {
      typename mesh_t::Point from, to, axis, proj;

      from = _m.point(_m.from_vertex_handle(_hh));
      to = _m.point(_m.to_vertex_handle(_hh));
      axis = to - from;

      if (axis.norm() != real_t(0.0))
      {
          axis.normalize();
          
          _pt = _pt - from;
          // project point on axis
          proj = axis * real_t((_pt | axis));

          // reflect on axis
          proj = (proj * real_t(2.0)) - _pt;
          _pt = _pt + from;
      }
      else
          _pt.vectorize(real_t(0.0));
  }

  void compute_midpoint(mesh_t& _m, const typename mesh_t::EdgeHandle& _eh) 
  {
#define V( X ) vector_cast< typename mesh_t::Normal >( X )

      typename mesh_t::Point& pos = _m.property(ep_pos_, _eh);
      typename mesh_t::Point pt, rpt;
      typename mesh_t::VertexHandle vh;
      typename mesh_t::HalfedgeHandle hh, hh_c, hh_cc;

      if (_m.is_boundary(_eh))
      {
          // ------ handle the 2 edge vertices ------
          hh = _m.halfedge_handle(_eh, 0);
          vh = _m.to_vertex_handle(hh);
          pt = _m.point(vh);
          
          vh = _m.from_vertex_handle(hh);
          pt = pt + _m.point(vh);
          
          pos = pt * real_t(9.0/16.0);

          // ------ handle the 2 adjacent vertices on the boundary ------
          hh_c = _m.cw_rotated_halfedge_handle(hh);
          if (!_m.is_boundary(hh_c))
              hh_c = _m.ccw_rotated_halfedge_handle(hh);
          vh = _m.to_vertex_handle(hh_c);
          pt = _m.point(vh);

          hh = _m.opposite_halfedge_handle(hh);
          hh_cc = _m.ccw_rotated_halfedge_handle(hh);
          if (!_m.is_boundary(hh_cc))
              hh_cc = _m.cw_rotated_halfedge_handle(hh);
          vh = _m.to_vertex_handle(hh_cc);
          pt = pt + _m.point(vh);
          
          pos = pos - (pt * real_t(1.0/16.0));
      }
      else
      {
          // ------ handle the 2 edge vertices ------
          hh = _m.halfedge_handle(_eh, 0);
          vh = _m.to_vertex_handle(hh);
          pt = _m.point(vh);
          
          vh = _m.from_vertex_handle(hh);
          pt = pt + _m.point(vh);
          
          pos = pt * real_t(0.5);
          
          // ------ handle the remaining 2 vertices of the 2 faces adjacent to the edge ------
          hh_c = _m.cw_rotated_halfedge_handle(hh);
          vh = _m.to_vertex_handle(hh_c);
          pt = _m.point(vh);
          
          hh_cc = _m.ccw_rotated_halfedge_handle(hh);
          vh = _m.to_vertex_handle(hh_cc);
          pt = pt + _m.point(vh);
          
          pos = pos + (real_t(1.0f/8.0f)*pt);
          
          // ------ handle the 4 vertices forming the butterfly shape over the considered edge ------
          if (_m.is_boundary(hh_c))
          {
              rpt = _m.point(_m.to_vertex_handle(hh));
              reflect_point(_m, rpt, hh_c);
              pt = rpt;
          }
          else
          {
              hh_c = _m.cw_rotated_halfedge_handle(hh_c);
              vh = _m.to_vertex_handle(hh_c);
              pt = _m.point(vh);
          }
          
          if (_m.is_boundary(hh_cc))
          {
              rpt = _m.point(_m.to_vertex_handle(hh));
              reflect_point(_m, rpt, hh_cc);
              pt = pt + rpt;
          }
          else
          {
              hh_cc = _m.ccw_rotated_halfedge_handle(hh_cc);
              vh = _m.to_vertex_handle(hh_cc);
              pt = pt + _m.point(vh);
          }
          
          hh = _m.opposite_halfedge_handle(hh);
          
          hh_c = _m.cw_rotated_halfedge_handle(hh);
          if (_m.is_boundary(hh_c))
          {
              rpt = _m.point(_m.to_vertex_handle(hh));
              reflect_point(_m, rpt, hh_c);
              pt = pt + rpt;
          }
          else
          {
              hh_c = _m.cw_rotated_halfedge_handle(hh_c);
              vh = _m.to_vertex_handle(hh_c);
              pt = pt + _m.point(vh);
          }
          
          hh_cc = _m.ccw_rotated_halfedge_handle(hh);
          if (_m.is_boundary(hh_cc))
          {
              rpt = _m.point(_m.to_vertex_handle(hh));
              reflect_point(_m, rpt, hh_cc);
              pt = pt + rpt;
          }
          else
          {
              hh_cc = _m.ccw_rotated_halfedge_handle(hh_cc);
              vh = _m.to_vertex_handle(hh_cc);
              pt = pt + _m.point(vh);
          }
          
          pos = pos - (real_t(1.0f/16.0f)*pt);
      }
#undef V
  }


  // compute midpoint for an edge with extraordinary vertex
  typename mesh_t::Point extra_midpoint(mesh_t& _m, const typename mesh_t::HalfedgeHandle& _hh, const int& _val)
  {
      typename mesh_t::Point pt;
      typename mesh_t::HalfedgeHandle hh_cc;
      int val_i = 0;
      std::vector<real_t>* weight_vec = NULL;

      if (_val > max_valence_)
      {
          // no precomputed weights available --> compute them now
          weights_.resize(_val-2);
          for(int i=(max_valence_+1); i <= _val; i++)
          {
              weights_[i-3].resize(i);
              compute_weight(weights_[i-3], i);
          }
          max_valence_ = _val;
      }

      // get vertex weight vector for corresponding valence number
      weight_vec = &weights_[(_val-3)];

      // weight the extraordinary point
      pt = _m.point(_m.from_vertex_handle(_hh)) * real_t(3.0/4.0);

      // weighted sum of 1-ring neighbors
      pt = pt + _m.point(_m.to_vertex_handle(_hh)) * (*weight_vec)[val_i];
      for(hh_cc = _m.ccw_rotated_halfedge_handle(_hh); hh_cc != _hh; hh_cc = _m.ccw_rotated_halfedge_handle(hh_cc))
      {
          val_i++;
          pt = pt + (_m.point(_m.to_vertex_handle(hh_cc)) * (*weight_vec)[val_i]);
      }

      return pt;
  }


private: // data

  OpenMesh::EPropHandleT< typename mesh_t::Point > ep_pos_;

  int           max_valence_;
  weights_t     weights_;

  ButterflyMode mode_;
};


//=============================================================================
#endif // BUTTERFLYT_H defined
//=============================================================================
