#include "decimater.h"

// qt
#include <qspinbox.h>
#include <qcheckbox.h>

std::string Decimater::draw_mode_ = "VEC:Collapsing Edges";

Decimater::Decimater(TriMesh& _mesh, QObject* _parent, const char* _name)
: Algorithm(_mesh, _parent, _name), decimater_(0), cur_iter_(-1)
{
    decimaterWdg_ = NULL;

    mod_roundness_ = false;
    mod_norm_flip_ = false;
    mod_ind_sets_ = false;

    last_n_frame_ = -1;
    last_frame_ = 0;
    no_animation_ = false;

    if (_parent != NULL)
    {
        // create and add decimater widget
        decimaterWdg_ = new DecimaterWidget(scrollView_->viewport(), "DecimaterWidget");
        scrollView_->addChild(decimaterWdg_);
        scrollView_->viewport()->setPaletteBackgroundColor(decimaterWdg_->paletteBackgroundColor());
        
        // choose settings for tool bar
        toolBar_->setCaption("Decimater");
        // check whether widget fits into dock area and adjust size
        QSize size = toolBar_->sizeHint();
        QDockArea* area = toolBar_->area();
        if (area != NULL)   // toolbar docked
        {
            int width;
            if ((area->height()-20) > decimaterWdg_->minimumHeight())
                width = (size.width() / 2) + decimaterWdg_->minimumWidth();
            else
                width = size.width() + scrollView_->sizeHint().width();
            toolBar_->setFixedExtentWidth(width);
        }
        
        // set control elements
        set_modeBox(decimaterWdg_->modeComboBox());
        set_slider(decimaterWdg_->frameSlider());
        set_playButton(decimaterWdg_->playPushButton());

        // get and set initial iteration count
        set_iter_count(decimaterWdg_->iterCountSpinBox()->value());

        // get and set initial edge collapse
        set_edge_collapse(decimaterWdg_->edgeCollapseSpinBox()->value());

        // get and set initial collapse constraints
        set_mod_roundness();
        set_mod_norm_flip();
        set_mod_ind_sets();
        
        connect(decimaterWdg_->modeComboBox(), SIGNAL(activated(int)), SLOT(set_mode(int)));
        connect(decimaterWdg_->iterCountSpinBox(), SIGNAL(valueChanged(int)), SLOT(set_iter_count(int)));
        connect(decimaterWdg_->edgeCollapseSpinBox(), SIGNAL(valueChanged(int)), SLOT(set_edge_collapse(int)));
        connect(decimaterWdg_->roundnessCheckBox(), SIGNAL(clicked()), SLOT(set_mod_roundness()));
        connect(decimaterWdg_->normFlipCheckBox(), SIGNAL(clicked()), SLOT(set_mod_norm_flip()));
        connect(decimaterWdg_->indSetsCheckBox(), SIGNAL(clicked()), SLOT(set_mod_ind_sets()));
        connect(decimaterWdg_->runPushButton(), SIGNAL(clicked()), SLOT(run()));
        connect(decimaterWdg_->playPushButton(), SIGNAL(pressed()), SLOT(animate()));
        connect(decimaterWdg_->frameSlider(), SIGNAL(valueChanged(int)), SLOT(set_frame(int)));

        disconnect(timer_, SIGNAL(timeout()), this, 0);
        connect(timer_, SIGNAL(timeout()), SLOT(delay_animate()));
        
        add_mode("Quadrics");
        add_mode("Roundness");
        
        // set first decimater mode
        if (n_modes_ >= 1)
            set_mode(1);
    }
}

Decimater::~Decimater()
{
    if(frame_ == last_n_frame_)
        decimate();

    for( ; !posVec_.empty(); posVec_.pop_back())
        mesh_.remove_property(posVec_.back());

    if (mesh_.has_vertex_texcoords())
    {
        for( ; !texVec_.empty(); texVec_.pop_back())
            mesh_.remove_property(texVec_.back());
    }

    delete decimater_;

    // decimater widget will be deleted automatically by parent object
}

const std::string& Decimater::draw_mode_name()
{
    return draw_mode_;
}

void Decimater::set_mode(int _id)
{
    // call base class function
    Algorithm::set_mode(_id);

    if (decimaterWdg_->modeComboBox()->text(_id) == "Roundness")
    {
        decimaterWdg_->roundnessCheckBox()->setChecked(false);
        decimaterWdg_->roundnessCheckBox()->setEnabled(false);
        set_mod_roundness();
    }
    else
        decimaterWdg_->roundnessCheckBox()->setEnabled(true);
}

void Decimater::set_edge_collapse(int _n)
{
    edge_collapse_ = _n;
}

void Decimater::set_frame(int _f)
{
    _f = std::min(n_frames(), std::max(0, _f));     // should lie in [0..n_frames()]
    
    if(_f != last_frame_)
    {
        last_frame_ = _f;
        if (slider_ != NULL)
            slider_->setValue(_f);
    }

    if (no_animation_)
    {
        Algorithm::set_frame(_f);
        no_animation_ = false;
    }
    else if (!timer_->isActive())
        delay_animate();
}

void Decimater::delay_animate(void)
{
    if (frame_ < last_frame_)
    {
        frame_++;
        timer_->start(1000 / steps_, true);
    }
    else if (frame_ > last_frame_)
    {
        frame_--;
        timer_->start(1000 / steps_, true);
    }

    Algorithm::apply();
    emit gl_changed();
    emit frame_changed(frame_);
}

void Decimater::animate(void)
{
    if (frame_ == n_frames())
        Algorithm::set_frame(0);

    if (slider_ != NULL)
            slider_->setValue(n_frames());
}

void Decimater::set_mod_roundness(void)
{
    mod_roundness_ = decimaterWdg_->roundnessCheckBox()->isChecked();
}

void Decimater::set_mod_norm_flip(void)
{
    mod_norm_flip_ = decimaterWdg_->normFlipCheckBox()->isChecked();
}

void Decimater::set_mod_ind_sets(void)
{
    mod_ind_sets_ = decimaterWdg_->indSetsCheckBox()->isChecked();
}


void Decimater::init()
{
    if(frame_ == last_n_frame_)
        decimate();

    last_n_frame_ = n_frames();
    no_animation_ = true;

    // delete all pos
    for( ; !posVec_.empty(); posVec_.pop_back())
        mesh_.remove_property(posVec_.back());

    // delete all texture coordinates
    if (mesh_.has_vertex_texcoords())
    {
        for( ; !texVec_.empty(); texVec_.pop_back())
            mesh_.remove_property(texVec_.back());
    }

    // add initial position
    mesh_.add_property(posInit_);
    posVec_.push_back(posInit_);

    // add initial texture coordinates
    if (mesh_.has_vertex_texcoords())
    {
        mesh_.add_property(texInit_);
        texVec_.push_back(texInit_);
    }

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        mesh_.status(vIt).set_locked(false);                // unlock vertices (will be set by some collapse modules)
        mesh_.property(posInit_, vIt) = mesh_.point(vIt);

        if (mesh_.has_vertex_texcoords())
            mesh_.property(texInit_, vIt) = mesh_.texcoord(vIt);
    }

    TriMesh::EdgeIter eIt, eEnd = mesh_.edges_end();
    for(eIt = mesh_.edges_begin(); eIt != eEnd; ++eIt)
    {
        mesh_.status(eIt).set_tagged(false);                // clear tagged status
    }

    if(decimater_)
    {
        decimater_->remove(modPM_);
        delete decimater_;
    }

    decimater_ = new DecimaterT(mesh_);
    decimater_->add(modPM_);                // module to keep track of edge collapses (has no other influence)

    const std::string& mode = current_mode_name();

    if(mode == "Quadrics")
        decimater_->add(ModQuadricT::Handle());
    else if(mode == "Roundness")
        decimater_->add(ModRoundnessT::Handle());

    //
    // roundness collapse constraint: (also used in non-binary form)
    // ------------------------------
    //
    // collapse is illegal if resulting triangles are highly degenerated
    //
    // metric: roundness (radius of triangle circumference divided by shortes edge length)
    //
    if (mod_roundness_)
    {
        OpenMesh::Decimater::ModHandleT<ModRoundnessT> modRH;
        decimater_->add(modRH);

        ModRoundnessT& mr = decimater_->module(modRH);
        mr.set_min_angle(15.0f, true);                  // set threshold
    }
    
    //
    // normal flipping:
    // ----------------
    //
    // collapse is illegal if triangle normals after collapse differ highly from
    // those before the collapse
    //
    // metric: angle between normals
    //
    if (mod_norm_flip_)
        decimater_->add(ModNormalFlippingT::Handle());

    //
    // independent sets:
    // -----------------
    //
    // after each collapse the one ring neighbors of the resulting vertex are locked
    // and will not participate in decimation anymore
    //
    if (mod_ind_sets_)
        decimater_->add(ModIndependentSetsT::Handle());

    // initialize decimater
    bool init = decimater_->initialize();
    assert(init);
}


void Decimater::exec()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    size_t n, n_vert;
    size_t collapsed, remaining;
    int *to_col_list;
    int to_collapse;
    int i;
    ModProgMesh::InfoList::size_type cur_index, start_index;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    n = mesh_.n_vertices();
    n_vert = n;

    // compute total number of requested collapses (for all iterations together)
    to_col_list = new int[iter_count_];
    to_collapse = 0;
    for(i=0; i < iter_count_; i++)
    {
        to_col_list[i] = int(0.01f * edge_collapse_ * n);
        to_collapse = to_collapse + to_col_list[i];
        n = n - to_col_list[i];
    }

    // do the collapse
    collapsed = decimater_->decimate(size_t(to_collapse));

    // get info about collapsed edges
    const ModProgMesh::InfoList& infolist = decimater_->module(modPM_).infolist();

    //decimater_->info(std::cout);

    //
    // simulate multiple iterations by decomposing decimation result into junks
    //
    n = n_vert;
    remaining = n_vert - collapsed;
    start_index = 0;
    for(i=0; i < iter_count_; i++)
    {
        std::cout << "Decimating from " << std::max(n, remaining) << " to ";
        n = n - to_col_list[i];
        std::cout << std::max(n, remaining) << " vertices.\n";

        VPropPoint pos1;
        VPropTexCoord tex1;

        mesh_.add_property(pos1);
        posVec_.push_back(pos1);

        if (mesh_.has_vertex_texcoords())
        {
            mesh_.add_property(tex1);
            texVec_.push_back(tex1);
        }

        // initialize new vertex position properties with the initial ones
        TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            mesh_.property(pos1, vIt) = mesh_.property(posInit_, vIt);

            if (mesh_.has_vertex_texcoords())
                mesh_.property(tex1, vIt) = mesh_.property(texInit_, vIt);
        }

        // take track of the edge collapses of an iteration with the vertex position properties
        start_index = start_index + to_col_list[i];
        cur_index = std::min(start_index, collapsed);
        while(cur_index > 0)
        {
            cur_index--;
            const ModProgMesh::Info& info = infolist[cur_index];

            // copy pos1 from v1 to v0
            mesh_.property(pos1, info.v0) = mesh_.property(pos1, info.v1);

            // copy tex1 from v1 to v0
            if (mesh_.has_vertex_texcoords())
                mesh_.property(tex1, info.v0) = mesh_.property(tex1, info.v1);


            if (i == (iter_count_-1))
            {
                // last round so all the collapsed edges will be traversed
                // --> split edges back to full mesh
                TriMesh::HalfedgeHandle hh;
                
                // undelete vertex
                mesh_.status(info.v0).set_deleted(false);
                // uncollapse vertex
                hh = mesh_.vertex_split(info.v0, info.v1, info.vl, info.vr);
                // tag edge for visualization
                mesh_.status(mesh_.edge_handle(hh)).set_tagged(true);
            }
        }
    }

    // delete old edges and faces
    mesh_.garbage_collection();

    // show information dialog if expected decimation rate was not reached
    // (!! do it after the garbage collection otherwise the redrawing of the
    //     inconsistent mesh will let the program crash !!)
    if (collapsed != to_collapse)
        QMessageBox::information(NULL, "Decimater information", "Decimation could not be preformed up to requested degree.\nPlease check the collapse constraints.");

    delete [] to_col_list;
}


void Decimater::apply(int _iter, float _alpha)
{
    if(_iter == iter_count_)
    {
        --_iter;
        _alpha = 1;
    }

    VPropPoint pos1 = posVec_[_iter + 1];
    VPropPoint pos0 = posVec_[_iter];

    VPropTexCoord tex0, tex1;
    if (mesh_.has_vertex_texcoords())
    {
        tex1 = texVec_[_iter + 1];
        tex0 = texVec_[_iter];
    }

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        mesh_.set_point(vIt, lerp(mesh_.property(pos0, vIt), mesh_.property(pos1, vIt), _alpha));
        
        if (mesh_.has_vertex_texcoords())
        {
            mesh_.set_texcoord(vIt, lerp(mesh_.property(tex0, vIt), mesh_.property(tex1, vIt), _alpha));
        }
    }

    // all properties need to be updated
    mesh_.update_later();
}

void Decimater::decimate()
{
    const ModProgMesh::InfoList& infolist = decimater_->module(modPM_).infolist();

    // collapse edges back to decimated mesh
    ModProgMesh::InfoList::const_iterator it, end = infolist.end();
    for(it = infolist.begin(); it != end; ++it)
        mesh_.collapse(mesh_.find_halfedge(it->v0, it->v1));

    // delete old edges and faces
    mesh_.garbage_collection();

    // all properties need to be updated
    mesh_.update_later();
}