#include "frequency.h"


std::string Frequency::draw_mode_ = "CV:Frequency Mass";


Frequency::Frequency(TriMesh& _mesh, QObject* _parent, const char* _name)
: Algorithm(_mesh, _parent, _name)
{
    mat_EV_ = NULL;
    vec_EW_ = NULL;
    mat_coeff_ = NULL;

    decomp_mode_ = eDecompPart;     // partial EW/EV-decomposition
    in_decomp_ = false;
    
    int n_vertices = mesh_.n_vertices();
    max_freq_num_ = n_vertices;
    cur_freq_idx_ = 0;
    freq_from_idx_ = 0;
    freq_to_idx_ = 0;

    frequencyWdg_ = NULL;

	mesh_.add_property(initPos_);                       // add property to store initial vertex positions
    mesh_.add_property(frequency, "Frequency Mass");    // property for frequency coloring
    mesh_.add_property(vIndex);                         // add vertex index property

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    int i = 0;
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
	    // assign index to vertex
        mesh_.property(vIndex, vIt) = i;
        i++;

        mesh_.property(initPos_, vIt) = mesh_.point(vIt);
    }

    if (_parent != NULL)
    {
        frequencyWdg_ = new FrequencyWidget(scrollView_->viewport(), "FrequencyWidget");
        scrollView_->addChild(frequencyWdg_);
        scrollView_->viewport()->setPaletteBackgroundColor(frequencyWdg_->paletteBackgroundColor());

        toolBar_->setCaption("Frequency");
        // check whether widget fits into dock area and adjust size
        QSize size = toolBar_->sizeHint();
        QDockArea* area = toolBar_->area();
        if (area != NULL)   // toolbar docked
        {
            int width;
            if ((area->height()-20) > frequencyWdg_->minimumHeight())
                width = (size.width() / 2) + frequencyWdg_->minimumWidth();
            else
                width = size.width() + scrollView_->sizeHint().width();
            toolBar_->setFixedExtentWidth(width);
        }
        
        QString str;
        str.sprintf("of %d", n_vertices);
        frequencyWdg_->decMaxFreqLabel()->setText(str);
        frequencyWdg_->decNumFreqSpinBox()->setMaxValue(std::max(0, (n_vertices-2)));    // maximal number of frequencies to compute
        if (n_vertices > 30)
            frequencyWdg_->decNumFreqSpinBox()->setValue(10);
        else if (n_vertices > 4)
            frequencyWdg_->decNumFreqSpinBox()->setValue(4);

        frequencyWdg_->decFreqSelComboBox()->insertItem("Largest (magnitude)", 0);
        frequencyWdg_->decFreqSelComboBox()->insertItem("Smallest (magnitude)", 1);
        frequencyWdg_->decFreqSelComboBox()->insertItem("Largest real part", 2);
        frequencyWdg_->decFreqSelComboBox()->insertItem("Smallest real part", 3);
        frequencyWdg_->decFreqSelComboBox()->setCurrentItem(1);
        set_dec_freq_sel(1);

        connect(frequencyWdg_->runPushButton(), SIGNAL(clicked()), SLOT(decompose()));
        connect(frequencyWdg_->resetPushButton(), SIGNAL(clicked()), SLOT(reset()));

        connect(frequencyWdg_->decButtonGroup(), SIGNAL(clicked(int)), SLOT(set_dec_mode(int)));
        connect(frequencyWdg_->decFreqSelComboBox(), SIGNAL(activated(int)), SLOT(set_dec_freq_sel(int)));

        connect(frequencyWdg_->freqSlider(), SIGNAL(valueChanged(int)), SLOT(set_cur_freq(void)));
        connect(frequencyWdg_->freqFromSlider(), SIGNAL(valueChanged(int)), SLOT(set_freq_from(int)));
        connect(frequencyWdg_->freqToSlider(), SIGNAL(valueChanged(int)), SLOT(set_freq_to(int)));
    }
}

Frequency::~Frequency(void)
{
    mesh_.remove_property(initPos_);
	mesh_.remove_property(frequency);
    mesh_.remove_property(vIndex);          // remove index property

    if (mat_EV_ != NULL)
        delete [] mat_EV_;

    if (vec_EW_ != NULL)
        delete [] vec_EW_;

    if (mat_coeff_ != NULL)
        delete [] mat_coeff_;

    // frequency widget will be deleted automatically by parent object
}

const std::string& Frequency::draw_mode_name()
{
	return draw_mode_;
}


/*======================================================================
  = restore old mesh                                                   =
  ======================================================================*/
void Frequency::reset(void)
{
    // set frequency selection sliders
    frequencyWdg_->freqFromSlider()->setValue(0);
    frequencyWdg_->freqFromSlider()->setMaxValue(0);
    frequencyWdg_->freqToSlider()->setValue(0);
    frequencyWdg_->freqToSlider()->setMaxValue(0);

    // set slider to number of eigenvectors
    frequencyWdg_->freqSlider()->setValue(0);
    frequencyWdg_->freqSlider()->setMaxValue(0);

    freqRangePaint();

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        mesh_.set_point(vIt, mesh_.property(initPos_, vIt));
    }

    // all properties need to be updated
	mesh_.update_later();

    emit gl_changed();
}


/*======================================================================
  = set decomposition mode                                             =
  ======================================================================*/
void Frequency::set_dec_mode(int id)
{
    bool enable = false;

    if (id == 0)
    {
        decomp_mode_ = eDecompFull;
        enable = false;
    }
    else if (id == 1)
    {
        decomp_mode_ = eDecompPart;
        enable = true;
    }

    // enable or disable additional options
    frequencyWdg_->decPartGroupBox()->setEnabled(enable);
    QObjectList *list = (QObjectList*)frequencyWdg_->decPartGroupBox()->children();
    QObject *obj;
    for (obj = list->first(); obj; obj = list->next())
    {
        ((QWidget*)obj)->setEnabled(enable);
    }
}


/*======================================================================
  = get number of eigenvalues to compute                               =
  ======================================================================*/
int Frequency::get_dec_freq_num(void)
{
    return frequencyWdg_->decNumFreqSpinBox()->value();
}


/*======================================================================
  = set which eigenvalues to compute                                   =
  ======================================================================*/
void Frequency::set_dec_freq_sel(int idx)
{
    switch(idx)
    {
        case 0:
            dec_freq_sel_[0] = 'L';  dec_freq_sel_[1] = 'M';
        break;
        case 1:
            dec_freq_sel_[0] = 'S';  dec_freq_sel_[1] = 'M';
        break;
        case 2:
            dec_freq_sel_[0] = 'L';  dec_freq_sel_[1] = 'R';
        break;
        case 3:
            dec_freq_sel_[0] = 'S';  dec_freq_sel_[1] = 'R';
        break;
    }
}


/*======================================================================
  = visualize range of selected frequencies (eigenvectors)             =
  ======================================================================*/
void Frequency::freqRangePaint(void)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    float width, height;
    float delta;
    int start, end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    QLabel *label = frequencyWdg_->freqPixmapLabel();

    QRect rect = label->rect();
    width = rect.width();
    height = rect.height();
    delta = width / (max_freq_num_-1);

    QPixmap pix(rect.size());
    pix.fill(label->backgroundColor());

    QPainter p(&pix);
    p.setBrush(blue);
    p.setPen(NoPen);

    start = delta * freq_from_idx_;
    end = delta * freq_to_idx_;
    if (freq_from_idx_ <= freq_to_idx_)
    {
        p.drawRect(start, 0, end - start, height);
    }
    else
    {
        p.drawRect(0, 0, end, height);
        p.drawRect(start, 0, width - start, height);
    }

    p.setPen(SolidLine);
    p.setPen(red);
    start = delta * cur_freq_idx_;
    if (start > (width - 5))            // marker should stay in visible area
        start = width - 5;
    p.drawLine(start, 0, start, height);

    label->setPixmap(pix);
}


/*======================================================================
  = check current frequency index and correct it if necessary          =
  ======================================================================*/
void Frequency::check_freq_idx()
{
    int idx = cur_freq_idx_;

    if (freq_from_idx_ <= freq_to_idx_)
    {
        if (idx < freq_from_idx_)
            frequencyWdg_->freqSlider()->setValue(freq_from_idx_);
        if (idx > freq_to_idx_)
            frequencyWdg_->freqSlider()->setValue(freq_to_idx_);
    }
    else
    {
        if ((idx < freq_from_idx_) && (idx > freq_to_idx_))
            frequencyWdg_->freqSlider()->setValue(freq_to_idx_);
    }
}


void Frequency::set_freq_from(int freq_idx)
{
    freq_from_idx_ = freq_idx;

    check_freq_idx();
    freqRangePaint();

    if (!in_decomp_)
        reconstruct();
}


void Frequency::set_freq_to(int freq_idx)
{
    freq_to_idx_ = freq_idx;

    check_freq_idx();
    freqRangePaint();

    if (!in_decomp_)
        reconstruct();
}


void Frequency::set_cur_freq(void)
{
    // get current frequency index from slider
    cur_freq_idx_ = frequencyWdg_->freqSlider()->value();
    freqRangePaint();

    set_freq_prop();
}


void Frequency::init_algo()
{
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
		// set frequency property value
        mesh_.property(frequency, vIt) = 0.0f;
    }

    int size = mesh_.n_vertices();
    int dim = 3;

    // free memory and reallocate
    if (mat_EV_ != NULL)
        delete [] mat_EV_;
    mat_EV_ = new double[size*max_freq_num_];

    // no size adaptation for partial decomposition
    if (vec_EW_ == NULL)
        vec_EW_ = new double[2*size];

    // free memory and reallocate
    if (mat_coeff_ != NULL)
        delete [] mat_coeff_;
    mat_coeff_ = new double[max_freq_num_*dim];
}


/*======================================================================
  = Set frequency property for mesh vertices                           =
  ======================================================================*/
void Frequency::set_freq_prop(void)
{
    int size = mesh_.n_vertices();
    int const_index = cur_freq_idx_*size;
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        int index = mesh_.property(vIndex, vIt);
        
        // get frequency value from user selected eigenvector
        float freq_value = mat_EV_[const_index + index];

        // set frequency property value
        mesh_.property(frequency, vIt) = freq_value;
    }

    // update mesh coloring
    mesh_.update_later(TriMesh::ColorBit);

    emit gl_changed();
}


/*======================================================================
  = Reconstruct mesh from Laplacian eigenvectors                       =
  =                                                                    =
  = recompute the mesh vertices as a linear combination of eigenvectors=
  = and coefficients                                                   =
  ======================================================================*/
void Frequency::reconstruct()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    doublereal *vertices, *coeff;
    integer size;
    integer dim = 3;
    doublereal alpha, beta;
    integer incx = 1;
    integer incy = 1;
    char uplo;                      // math function settings

    int start_idx, end_idx;

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    size = mesh_.n_vertices();
    if (size < 1)
        return;

    vertices = new doublereal[size * dim];
    coeff = new doublereal[dim];

    uplo = ' ';
    alpha = 0.0;    // value for off diagonal elements
    beta = 0.0;     // value for diagonal elements
    dlaset_(&uplo, &size, &dim, &alpha, &beta, vertices, &size);  // set matrix entries to zero

    alpha = 1.0;
    if (freq_from_idx_ > freq_to_idx_)
    {   // use low and high frequencies but neglect some in between
        start_idx = 0;
        end_idx = freq_to_idx_;

        for(int i=start_idx; i <= end_idx; i++)
        {
            coeff[0] = mat_coeff_[i];
            coeff[1] = mat_coeff_[i + max_freq_num_];
            coeff[2] = mat_coeff_[i + (max_freq_num_ << 1)];

            // add reconstruction part of eigenvector
            //
            // Rank one update:  A = alpha*x*y + A
            dger_(&size, &dim, &alpha, &mat_EV_[i*size], &incx, coeff, &incy, vertices, &size);
        }

        start_idx = freq_from_idx_;
        end_idx = (max_freq_num_-1);
    }
    else
    {
        start_idx = freq_from_idx_;
        end_idx = freq_to_idx_;
    }

    for(int i=start_idx; i <= end_idx; i++)
    {
        coeff[0] = mat_coeff_[i];
        coeff[1] = mat_coeff_[i + max_freq_num_];
        coeff[2] = mat_coeff_[i + (max_freq_num_ << 1)];

        // add reconstruction part of eigenvector
        //
        // Rank one update:  A = alpha*x*y + A
        dger_(&size, &dim, &alpha, &mat_EV_[i*size], &incx, coeff, &incy, vertices, &size);
    }

    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        int index = mesh_.property(vIndex, vIt);

        TriMesh::Point tmpPt;
        tmpPt[0] = vertices[index];
        tmpPt[1] = vertices[index + size];
        tmpPt[2] = vertices[index + (size << 1)];

		mesh_.set_point(vIt, tmpPt);
    }

    delete [] vertices;
    delete [] coeff;

    // all properties need to be updated
	mesh_.update_later();

    emit gl_changed();
}


/*======================================================================
  = full eigenvalue/eigenvector decomposition of matrix A              =
  =                                                                    =
  = Note: Matrices are stored column wise                              =
  =                                                                    =
  = Parameters:                                                        =
  =     [in]    size        problem size                               =
  =     [in/out]A           matrix to decompose (will be written to)   =
  =     [out]   EW          (size,2)-matrix of (complex) eigenvalues   =
  =     [out]   EV          eigenvector matrix                         =
  ======================================================================*/
int Frequency::eigen_decomp(int size, double *A, double *EW, double *EV)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    char jobvl = 'N';                   // do not compute the left eigenvectors
    char jobvr = 'V';                   // compute the right eigenvectors
    integer n = size;                   // matrix size
    integer lda = size;                 // leading dimension of matrix
    doublereal VL;                      // left eigenvectors
    integer ldvl = 1;                   // leading dimension of left eigenvector matrix
    integer ldvr = size;                // leading dimension of right eigenvector matrix
    integer lwork = 2*4*size;           // size of workspace array
    doublereal *work = new doublereal[lwork];   // workspace array
    integer info;                       // status value
    integer* idx = new integer[size];   // permutation index array
    logical forwrd = TRUE_;             // forward permutation
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//


    // eigenvalue/eigenvector decomposition (passed matrix will be overwritten!)
    dgeev_(&jobvl, &jobvr, &n, A, &lda, EW, &EW[size], &VL, &ldvl, EV, &ldvr, work, &lwork, &info);

    for(int i=1; i <= size; i++)
    {
        idx[i-1] = i;
    }
    quicksort(idx, EW, size);               // sort eigenvalues and generate permutation index
    // sort eigenvectors acording to eigenvalues
    dlapmt_(&forwrd, &n, &n, EV, &lda, idx);

    /*char uplo = 'U';                        // input matrix stores values in upper triangular part
    lwork = 4*size*size;                    // size of workspace array
    double *work = new double[lwork];       // workspace array
    integer liwork = 6*size;                // size of workspace array
    integer* iwork = new integer[liwork];   // workspace array

    memcpy(EV, A, size*size*sizeof(double));
    dsyevd_(&jobvr, &uplo, &n, EV, &n, EW, work, &lwork, iwork, &liwork, &info);*/

    delete [] work;
    delete [] idx;
    //delete [] iwork;    //!!!

    /*integer idigit = 3;         // number of digits to display
    char ifmt1[] = "Resulting Eigenvalues:";
    dvout_((integer*)stdout, &n, EW, &idigit, ifmt1, (ftnlen)22);*/

    return info;
}


/*======================================================================
  = get approximate eigenvalues and eigenvectors using the arnoldi     =
  = package                                                            =
  =                                                                    =
  = size   :    size of eigenproblem                                   =
  = A      :    matrix                                                 =
  = num_ev :    number of expected eigenvalues to compute              =
  ======================================================================*/
void Frequency::arnoldi(int size, SparseMatrix& A, int num_ev)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    integer ido = 0;                    // reverse communication flag
    integer n = size;                   // problem size
    integer nev = num_ev;               // number of expected eigenvalues to compute
    integer ncv = std::min(n, nev*5);   // number of Lanczos basis vectors to use
    integer lworkl = 3 * ncv * (ncv+6); // size of workspace
    integer iparam[11];
    integer ipntr[14];
    iparam[0] = 1;                      // use exact shifts
    iparam[2] = 500;                    // max number of IRLM iterations
    iparam[6] = 1;                      // use simple matrix-vector product
    integer info = 0;                   // error indicator
    char bmat = 'I';                    // use identity matrix for inner product
    char which[2];                      // selects which eigenvalues to compute ("LM", "SM", "LR", "SR", "LI", "SI")
    which[0] = dec_freq_sel_[0];  which[1] = dec_freq_sel_[1];
    doublereal tol = 0.;                // stopping criterion (default: machine epsilon)
    doublereal *resid = new doublereal[n];  // initial residual vector (random if ido = 0)
    doublereal *v = new doublereal[n*ncv];  // space for Lanczos vectors
    doublereal *workd = new doublereal[n*3];// work array
    doublereal *workl = new doublereal[lworkl]; // work space
    doublereal *workev = new doublereal[3*ncv];  // work space
    ftnlen bmat_len = 1;                // char length of bmat
    ftnlen which_len = 2;               // char length of which

    logical rvec = TRUE_;               // compute Ritz vectors
    char howmny = 'A';                  // compute all expected Ritz vectors
    logical *select = new logical[ncv]; // select Ritz vectors to compute
    doublereal* rval_r = new doublereal[nev+1]; // real part of Ritz values
    doublereal* rval_i = new doublereal[nev+1]; // imaginary part of Ritz values
    doublereal* rv = new doublereal[n*(nev+1)]; // Ritz vectors
    doublereal sigma_r;                 // not used in this mode
    doublereal sigma_i;                 // not used in this mode
    ftnlen howmny_len = 1;              // char length of howmny
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    // 
    // run arnoldi/lanczos itertion
    //
    dnaupd_(&ido, &bmat, &n, which, &nev, &tol, resid, &ncv, v, &n, iparam, ipntr, workd, workl, &lworkl, &info, bmat_len, which_len);
    while ((ido == -1) || (ido == 1))
    {
        A.mul_vec(size, &workd[ipntr[0]-1], &workd[ipntr[1]-1]);

        dnaupd_(&ido, &bmat, &n, which, &nev, &tol, resid, &ncv, v, &n, iparam, ipntr, workd, workl, &lworkl, &info, bmat_len, which_len);
    }

    if (info != 0)
        printf("\nLanczos iteration error flag was set:  flag = %d\n", info);

    //
    // postprocessing to get EWs/EVs
    //
    dneupd_(&rvec, &howmny, select, rval_r, rval_i, rv, &n, &sigma_r, &sigma_i, workev, &bmat, &n, which, &nev, &tol, 
	        resid, &ncv, v, &n, iparam, ipntr, workd, workl, &lworkl, &info, howmny, bmat_len, which_len);

    // store computed aprroximations of EWs/EVs
    memcpy(vec_EW_, rval_r, nev*sizeof(double));

    memcpy(mat_EV_, rv, n*nev*sizeof(double));
    
    integer idigit = 3;         // number of digits to display
    integer ld = 1;
    char ifmt1[] = "Resulting Ritz Values:";
    dmout_((integer*)stdout, &nev, &ld, rval_r, &nev, &idigit, ifmt1, (ftnlen)22);

    delete [] resid;
    delete [] v;
    delete [] workd;
    delete [] workl;
    delete [] workev;
    delete [] select;
    delete [] rval_r;
    delete [] rval_i;
    delete [] rv;
}


/*======================================================================
  = Decompose mesh into frequencies                                    =
  ======================================================================*/
void Frequency::decompose()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    integer dim = 3;
    integer size;                   // size for vectors/matrices
    integer ldc;
    double *mat_x = NULL;           // mesh vertex matrix
    double *mat_Lap = NULL;         // discretized Laplacian
    double alpha, beta;
    char transa, transb, uplo;      // math function settings

    SparseMatrix sp_mat_Lap;        // discretized Laplacian stored in a
                                    // data structure for sparse matrices

    QTime time;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//


    size = mesh_.n_vertices();

    if (size < 1)
        return;

    // set number of eigenvectors to compute
    if (decomp_mode_ == eDecompFull)
        max_freq_num_ = size;
    else
        max_freq_num_ = get_dec_freq_num();

    init_algo();

    time.start();                                           // do time measurement

    TriMesh::ConstVertexVertexIter cvvIt;                   // vertex neighborhood iterator
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    in_decomp_ = true;
    
    mat_x = new double[size*dim];                           // mesh vertex matrix

    if (decomp_mode_ == eDecompFull)
    {
        mat_Lap = new double[size*size];                    // discretized Laplacian
        uplo = ' ';
        alpha = 0.;     // value for off diagonal elements
        beta = 1.;      // value for diagonal elements
        dlaset_(&uplo, &size, &size, &alpha, &beta, mat_Lap, &size);   // set matrix identity
    }
    
	// go through all mesh vertices and compute discretization of Laplacian
    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        int index = mesh_.property(vIndex, vIt);            // get vertex index
        TriMesh::Point pt = mesh_.point(vIt);               // get vertex coordinates
        // insert vertex into mesh vertex matrix
        mat_x[index] = pt[0];
        mat_x[index + size] = pt[1];
        mat_x[index + (size << 1)] = pt[2];

        // compute valence (number of neighbors)
        int valence = 0;
		for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
        {
			valence++;
		}

        if (valence != 0)
        {
            if (decomp_mode_ == eDecompPart)
            {
                // set the diagonal matrix entry
                sp_mat_Lap.insert(index, index, 1.0);
            }

            float factor = -1.0f / (float)valence;
            for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
            {
			    int tmpIndex = mesh_.property(vIndex, cvvIt);

                if (decomp_mode_ == eDecompFull)
                {
                    // set matrix entry for this neighbor vertex
                    mat_Lap[index + tmpIndex*size] = factor;
                }
                else if (decomp_mode_ == eDecompPart)
                {
                    // set matrix entry
                    sp_mat_Lap.insert(index, tmpIndex, factor);
                }
		    }
        }
    }

    if (decomp_mode_ == eDecompFull)
    {
        // do the eigenvalue/eigenvector decomposition of the Laplacian (passed matrix will be modified!)
        eigen_decomp(size, mat_Lap, vec_EW_, mat_EV_);
    }
    else if (decomp_mode_ == eDecompPart)
    {
        arnoldi(size, sp_mat_Lap, max_freq_num_);
    }

    // compute the coefficients for the linear combination of the eigenvectors
    // -----------------------------------------------------------------------
    // x = sum(i=1..N)( a(i) * ev(i) )
    // -----------------------------------------------------------------------
    // x     : vertex matrix (with columns x,y,z)
    // a(i)  : i-th coefficient vector
    // ev(i) : i-th eigenvector
    // --> a(i)(x) = <x[x],ev(i)>;  a(i)(y) = <x[y],ev(i)>;  a(i)(z) = <x[z],ev(i)>
    // (multiply transposed eigenvector matrix with mesh vertex matrix)
    transa = 'T';           // use transposed matrix
    transb = 'N';
    alpha = 1.0;
    beta = 0.0;
    ldc = max_freq_num_;
    dgemm_(&transa, &transb, &ldc, &dim, &size, &alpha, mat_EV_, &size, mat_x, &size, &beta, mat_coeff_, &ldc);

    printf("Computation time for decomposition (%d of %d eigenvalues):\t%.2e ms\n", max_freq_num_, size, (double)time.elapsed());     // show cumputation time

    // set frequency selection sliders
    frequencyWdg_->freqFromSlider()->setMaxValue((max_freq_num_-1));
    frequencyWdg_->freqFromSlider()->setValue(0);
    frequencyWdg_->freqToSlider()->setMaxValue((max_freq_num_-1));
    frequencyWdg_->freqToSlider()->setValue((size-1));

    // recompute the mesh vertices
    reconstruct();

    // set slider to number of eigenvectors
    frequencyWdg_->freqSlider()->setMaxValue((max_freq_num_-1));
    frequencyWdg_->freqSlider()->setValue(0);

    set_freq_prop();

    // free space
    delete [] mat_x;
    
    if (decomp_mode_ == eDecompFull)
        delete [] mat_Lap;

    in_decomp_ = false;
}
