#pragma once

// fortran to c
#include <f2c.h>

// blas
#include <blas.h>

// lapack
#include <clapack.h>

// arpack
#include <arpack.h>

#include <limits>
#include <vector>

#include "algorithm.h"
#include "frequencywidget.h"
#include "sparsematrix.h"
#include "sort.h"

// qt
#include <qframe.h>
#include <qlabel.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qspinbox.h>
#include <qslider.h>
#include <qgroupbox.h>
#include <qbuttongroup.h>
#include <qobjectlist.h>
#include <qdatetime.h>


class Frequency : public Algorithm
{
    Q_OBJECT

    // EW/EV-decomposition method
    enum enumDecompMode
    {
        eDecompFull,        // full decomposition
        eDecompPart,        // partial decomposition using arnoldi/lanczos iteration
    };

    typedef OpenMesh::VPropHandleT<TriMesh::Point> VPropPoint;
	typedef OpenMesh::VPropHandleT<float>          VPropFloat;
    typedef OpenMesh::VPropHandleT<int>            VPropInt;

public:
	Frequency(TriMesh& _mesh, QObject* _parent = 0, const char* _name = 0);
	virtual ~Frequency();

public:
	virtual const std::string& draw_mode_name();
    
public slots:
    void decompose(void);               // decomposition in frequencies
    void reset(void);
    void set_dec_mode(int id);          // set the decomposition mode
    void set_dec_freq_sel(int idx);     // set which eigenvalues to compute
    void set_freq_from(int freq_idx);
    void set_freq_to(int freq_idx);
    void set_cur_freq(void);

private:
    void init_algo();
	void reconstruct(void);             // reconstruct mesh from Laplacian eigenvectors
    void arnoldi(int size, SparseMatrix& A, int num_ev);    // arnoldi/lanczos iteration
    inline int eigen_decomp(int size, double *A, double *EW, double *EV);    // compute all EVs/EWs
    void check_freq_idx(void);          // check current frequency index and correct it if necessary
    void set_freq_prop(void);           // set frequency property for every mesh vertex
    int get_dec_freq_num(void);         // get number of eigenvalues to compute
    void freqRangePaint(void);          // visualize selection of frequencies

	static std::string draw_mode_;
    enumDecompMode decomp_mode_;

	VPropPoint initPos_;                // store initial positions
    VPropFloat frequency;
    VPropInt vIndex;

    FrequencyWidget *frequencyWdg_;     // tool widget for frequency

    bool in_decomp_;                    // set true during EV/EW decomposition
    
    char dec_freq_sel_[2];              // select which eigenvalues to compute (arnoldi/lanczos iteration)
    int max_freq_num_;                  // number of computed eigenvectors
    int cur_freq_idx_;                  // index of current eigenvector
    int freq_from_idx_, freq_to_idx_;   // index range for mesh reconstruction from eigenvectors

    double *mat_EV_;                    // mesh frequencies as eigenvectors of the Laplacian
    double *vec_EW_;                    // eigenvalues of the Laplacian
    double *mat_coeff_;                 // coefficients for linear combination of eigenvectors
};
