// stl
#include <iostream>
#include <fstream>
// qt
#include <qapplication.h>
#include <qfileinfo.h>
#include <qstringlist.h>
#include <qmainwindow.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qsignalmapper.h>
#include <qdir.h>
// openmesh
#include <OpenMesh/Core/IO/BinaryHelper.hh>
#include <OpenMesh/Core/Utils/Endian.hh>
#include <OpenMesh/Tools/Utils/Timer.hh>
//
#include "renderwidget.h"

std::string RenderWidget::nomode_ = "";

RenderWidget::RenderWidget(QWidget* _parent, const char* _name)
: MeshViewerWidget(_parent, _name), show_tedges_(false)
{
    tex_file_ = "";
    render_menu_ = NULL;

    init();
}

RenderWidget::~RenderWidget()
{
}

void RenderWidget::draw_scene(const std::string& _draw_mode)
{
	int bits = 0;

	if( _draw_mode == "Solid Flat" ||
		_draw_mode == "Solid Smooth" ||
		_draw_mode == "Strips'n VertexArrays" ||
		_draw_mode == "Smooth Colored Faces" ||
		show_vnormals_ || show_fnormals_)
		bits |= TriMesh::NormalBit;

	if( _draw_mode == "Colored Vertices" )
		bits |= TriMesh::ColorBit;

	// update face normal base
	if(show_fnormals_ && (mesh_.update_bits() & TriMesh::NormalBit))
	{
		TriMesh::FaceVertexIter fvIt;
		TriMesh::FaceIter fIt, fEnd = mesh_.faces_end();
		for (fIt = mesh_.faces_begin(); fIt != fEnd; ++fIt)
		{
			TriMesh::Point& v = mesh_.property( fp_normal_base_, fIt ).vectorize(0);
			for(fvIt = mesh_.fv_iter(fIt); fvIt; ++fvIt)
				v += mesh_.point(fvIt);
			v *= 1.0f/3;
		}
	}

	mesh_.update_properties(bits);

	Base::draw_scene(_draw_mode);

	if (show_tedges_)
	{
		glDisable(GL_LIGHTING);
		glDepthRange(0.0, 0.9999);
		glBegin(GL_LINES);
        GLfloat cur_color[4];
        glGetFloatv(GL_CURRENT_COLOR, cur_color);   // store current color
		glColor3f(0.282f, 1.0f, 0.702f);            // turqoise
		TriMesh::EdgeIter eIt, eEnd = mesh_.edges_end();
		for(eIt = mesh_.edges_begin(); eIt != eEnd; ++eIt)
		{
			if(mesh_.status(eIt).tagged())
			{
				TriMesh::HalfedgeHandle hh = mesh_.halfedge_handle(eIt, 0);
				glVertex( mesh_.from_vertex_handle(hh) );
				glVertex( mesh_.to_vertex_handle(hh) );
			}
		}
        glColor3fv(cur_color);                      // restore old color
		glEnd();
		glDepthRange(0.0, 1.0);
	}

}

void RenderWidget::init(void)
{
	render_menu_ = dynamic_cast<QPopupMenu*>(child("Draw Mode Menu", "QPopupMenu", 0));

	// add menu to menubar
	dynamic_cast<QMainWindow*>(parent())->menuBar()->insertItem("Rendering", render_menu_, -1, 2);

    // add draw mode for texturing
    add_draw_mode("Textured");
    
    // add draw mode, get id and remove again
	add_draw_mode("Colored Vertices");
	int cv_id = 1;
	while(render_menu_->text(cv_id) != "Colored Vertices")
		++cv_id;
	del_draw_mode("Colored Vertices");

	// add colored vertices submenu
	cv_submenu_ = new QPopupMenu(render_menu_, "Colored Vertices Submenu");
	connect(cv_submenu_, SIGNAL(activated(int)), SLOT(slotCvSubmenu(int)));    
	render_menu_->insertItem("Colored Vertices", cv_submenu_, cv_id);

	// emit popup_menu signal when submenu item has been selected
	QSignalMapper* mapper = new QSignalMapper(cv_submenu_, "cvSubmenuMapper");
	mapper->setMapping(cv_submenu_, cv_id);
	connect(cv_submenu_, SIGNAL(activated(int)), mapper, SLOT(map()));    
	connect(mapper, SIGNAL(mapped(int)), render_menu_, SIGNAL(activated(int)));    

	n_cv_modes_ = 0;
	add_cv_mode("Normal");
	add_cv_mode("1st Principle Curvature");
	add_cv_mode("2nd Principle Curvature");
	add_cv_mode("Mean Curvature");
	add_cv_mode("Gaussian Curvature");
	add_cv_mode("Shape Index");

	// add draw vectos submenu
	vec_submenu_ = new QPopupMenu(render_menu_, "Draw Vectors Submenu");
	connect(vec_submenu_, SIGNAL(activated(int)), SLOT(slotVecSubmenu(int)));    
	render_menu_->insertItem("Draw Vectors", vec_submenu_);

	n_vec_modes_ = 0;
	add_vec_mode("Face Normals");
	add_vec_mode("Vertex Normals");
}

bool RenderWidget::open_mesh(const char* _filename, OpenMesh::IO::Options _opt)
{
    if (tex_file_.empty())
    {
        // try to set standard texture file
        QString s = QDir::currentDirPath() + "/data/chess.png";
        set_texture_file(s);
    }

	if(!Base::open_mesh(_filename, _opt))
		return false;

	if(!mesh_.has_vertex_colors())
		mesh_.request_vertex_colors();

    if (!_opt.check(IO::Options::VertexTexCoord))
    {
        use_texture_ = true;
        mesh_.init_vertex_texcoords();      // define texture coordinates
    }

	// need to update curvature and color property
	mesh_.update_later(TriMesh::CurvatureBit | TriMesh::ColorBit);

    updateGL();

	return true;
}


/*======================================================================
  = set and load texture file                                          =
  ======================================================================*/
void RenderWidget::set_texture_file(const char* _filename)
{
    if (open_texture(_filename))
        tex_file_ = _filename;
}


/*======================================================================
  = enable/disable texture menu entry                                  =
  ======================================================================*/
void RenderWidget::set_texture_menu(bool enable)
{
    for(int idx=0; size_t(idx) < render_menu_->count(); ++idx)
    {
        int id = render_menu_->idAt(idx);
        if ((id != -1) && (render_menu_->text(id) == "Textured"))
        {
            // item should be disabled but is currently active
            if ((!enable) && (render_menu_->isItemChecked(id)))
            {
                if (render_menu_->count() > 0)
                {
                    render_menu_->activateItemAt(1);
                    updateGL();
                }
            }

            render_menu_->setItemEnabled(id, enable);
            break;
        }
    }
}


void RenderWidget::add_cv_mode(const std::string& _s)
{
	++n_cv_modes_;
	cv_submenu_->insertItem(_s.c_str(), n_cv_modes_);
	cv_mode_names_.push_back(_s);
}

void RenderWidget::del_cv_mode(const std::string& _s)
{
	std::vector<std::string>::iterator it =
		std::find(cv_mode_names_.begin(), cv_mode_names_.end(), _s );
  
#if _DEBUG
	assert( it != cv_mode_names_.end() );
#else
	if ( it == cv_mode_names_.end() )
		return;
#endif

	for(int idx=0; size_t(idx) < cv_submenu_->count(); ++idx)
	{
		int id = cv_submenu_->idAt(idx);
		if (id != -1 && cv_submenu_->text(id) == _s.c_str())
		{
            // submenu entry to delete is active
            if (cv_submenu_->isItemChecked(id))
            {
                cv_submenu_->setItemChecked(id, false);

                if (render_menu_->count() > 0)
                    render_menu_->activateItemAt(1);
            }

			cv_submenu_->removeItemAt(idx);
			break;
		}
	}
}

void RenderWidget::add_vec_mode(const std::string& _s)
{
	++n_vec_modes_;
	vec_submenu_->insertItem(_s.c_str(), n_vec_modes_);
	vec_mode_names_.push_back(_s);
}

void RenderWidget::del_vec_mode(const std::string& _s)
{
	std::vector<std::string>::iterator it =
		std::find(vec_mode_names_.begin(), vec_mode_names_.end(), _s );
  
#if _DEBUG
	assert( it != vec_mode_names_.end() );
#else
	if ( it == vec_mode_names_.end() )
		return;
#endif

	for(int idx=0; size_t(idx) < vec_submenu_->count(); ++idx)
	{
		int id = vec_submenu_->idAt(idx);
		if (id != -1 && vec_submenu_->text(id) == _s.c_str())
		{
            // deactivate visualization of collapsing edges
            if(strcmp(_s.c_str(), "Collapsing Edges") == 0)
		        show_tedges_ = false;

			vec_submenu_->removeItemAt(idx);
			break;
		}
	}
}

void RenderWidget::slotCvSubmenu(int _id)
{
	for (size_t i=1; i <= n_cv_modes_; ++i)
		cv_submenu_->setItemChecked(i, false);
	cv_mode_ = _id;
	cv_submenu_->setItemChecked(_id, true);

	const std::string& mode = current_cv_mode();

	if (mode == "Smoothing Distance")
	{
		OpenMesh::VPropHandleT<float> vph;
		mesh_.get_property_handle(vph, "Smoothing Distance");
		mesh_.set_cv_mode(vph);
	}

	else if (mode == "Subdivision Vertices")
	{
		OpenMesh::VPropHandleT<float> vph;
		mesh_.get_property_handle(vph, "Subdivision Vertices");
		mesh_.set_cv_mode(vph);
	}

    else if (mode == "Frequency Mass")
	{
		OpenMesh::VPropHandleT<float> vph;
		mesh_.get_property_handle(vph, "Frequency Mass");
		mesh_.set_cv_mode(vph);
	}

	else
	{
		mesh_.set_cv_mode(mode);
	}

}

void RenderWidget::slotVecSubmenu(int _id)
{
	bool on = !vec_submenu_->isItemChecked(_id);
	vec_submenu_->setItemChecked(_id, on);

	const std::string& mode = _id ? vec_mode_names_[_id-1] : nomode_;

	if(mode == "Face Normals")
		show_fnormals_ = on;
	else if(mode == "Vertex Normals")
		show_vnormals_ = on;
	else if(mode == "Collapsing Edges")
		show_tedges_ = on;

	mesh_.update_later(TriMesh::NormalBit);
}
