#include "smoother.h"

std::string Smoother::draw_mode_ = "CV:Smoothing Distance";

Smoother::Smoother(TriMesh& _mesh, QObject* _parent, const char* _name)
: Algorithm(_mesh, _parent, _name), cur_iter_(-1)
{
    mesh_.add_property(distances, "Smoothing Distance");

    // add vertex index property
    mesh_.add_property(vIndex);

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    int i = 0;
    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        // assign index to vertex
        mesh_.property(vIndex, vIt) = i;
        i++;
    }

    int_mode_ = eIntExpl;           // explicit integration
    min_int_step_ = std::numeric_limits<float>::epsilon();
    max_int_step_ = 100.0f;
    int_step_size_ = 1.0f;          // integration step size
    vol_pres_ = false;              // volume preservation

    smootherWdg_ = NULL;

    if (_parent != NULL)
    {
        // create and add smoother widget
        smootherWdg_ = new SmootherWidget(scrollView_->viewport(), "SmootherWidget");
        scrollView_->addChild(smootherWdg_);
        scrollView_->viewport()->setPaletteBackgroundColor(smootherWdg_->paletteBackgroundColor());
        
        // choose settings for tool bar
        toolBar_->setCaption("Smoother");
        // check whether widget fits into dock area and adjust size
        QSize size = toolBar_->sizeHint();
        QDockArea* area = toolBar_->area();
        if (area != NULL)   // toolbar docked
        {
            int width;
            if ((area->height()-20) > smootherWdg_->minimumHeight())
                width = (size.width() / 2) + smootherWdg_->minimumWidth();
            else
                width = size.width() + scrollView_->sizeHint().width();
            toolBar_->setFixedExtentWidth(width);
        }
        
        // set control elements
        set_modeBox(smootherWdg_->modeComboBox());
        set_slider(smootherWdg_->frameSlider());
        set_playButton(smootherWdg_->playPushButton());

        // set validator range for step size input
        smootherWdg_->setValidatorRange(min_int_step_, max_int_step_, 10);
        
        connect(smootherWdg_->modeComboBox(), SIGNAL(activated(int)), SLOT(set_mode(int)));
        connect(smootherWdg_->iterCountSpinBox(), SIGNAL(valueChanged(int)), SLOT(set_iter_count(int)));
        connect(smootherWdg_->intButtonGroup(), SIGNAL(clicked(int)), SLOT(set_int_mode(int)));
        connect(smootherWdg_->volPresCheckBox(), SIGNAL(clicked()), SLOT(set_vol_pres()));
        connect(smootherWdg_->runPushButton(), SIGNAL(clicked()), SLOT(run()));
        connect(smootherWdg_->playPushButton(), SIGNAL(pressed()), SLOT(animate()));
        connect(smootherWdg_->frameSlider(), SIGNAL(valueChanged(int)), SLOT(set_frame(int)));
        
        // set the different smoothing modes
        add_mode("Umbrella");
        add_mode("Improved Umbrella");
        add_mode("Curvature Flow");
        add_mode("2nd Order Difference");

        // set first smoothing mode
        if (n_modes_ >= 1)
            set_mode(1);
    }
}

Smoother::~Smoother(void)
{
    for( ; !posVec.empty(); posVec.pop_back())
        mesh_.remove_property(posVec.back());

    mesh_.remove_property(distances);

    // remove index property
    mesh_.remove_property(vIndex);

    // smoother widget will be deleted automatically by parent object
}

const std::string& Smoother::draw_mode_name()
{
    return draw_mode_;
}

/*======================================================================
  = set integration mode                                               =
  ======================================================================*/
void Smoother::set_int_mode(int id)
{
    if (id == 0)
    {
        int_mode_ = eIntExpl;
    }
    else
    {
        int_mode_ = eIntImpl;
    }
}

/*======================================================================
  = set integration step size                                          =
  ======================================================================*/
bool Smoother::set_int_step_size(void)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    double value;
    QString text;
    bool valid = false;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    text = smootherWdg_->intStepSizeLineEdit()->text();
    value = text.toDouble(&valid);
        
    if (valid)
        int_step_size_ = value;

    return valid;
}

/*======================================================================
  = enable/disable volume preservation                                 =
  ======================================================================*/
void Smoother::set_vol_pres(void)
{
    vol_pres_ = smootherWdg_->volPresCheckBox_->isChecked();
}

void Smoother::init()
{
    // delete all pos
    for( ; !posVec.empty(); posVec.pop_back())
        mesh_.remove_property(posVec.back());

    // add initial position
    VPropPoint pos0;
    mesh_.add_property(pos0);
    posVec.push_back(pos0);

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        mesh_.property(pos0, vIt) = mesh_.point(vIt);
}

/*======================================================================
  = calculate volume of mesh (for volume preservation)                 =
  ======================================================================*/
float Smoother::mesh_volume(VPropPoint *pos)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    TriMesh::ConstFaceVertexIter cfvIt;
    TriMesh::FaceIter fIt, fEnd = mesh_.faces_end();

    float volume = 0.0f;
    TriMesh::Point pt[3];
    TriMesh::Point ptm, ptn;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    for(fIt = mesh_.faces_begin(); fIt != fEnd; ++fIt)
    {
        cfvIt = mesh_.fv_iter(fIt);

        // get triangle vertices
        for(int i=0; i < 3; i++)
        {
            assert(cfvIt != NULL);

            pt[i] = mesh_.property(*pos, cfvIt);
            ++cfvIt;
        }

        ptm = (1.0f / 3.0f) * (pt[0] + pt[1] + pt[2]);
        ptn = cross((pt[1] - pt[0]), (pt[2] - pt[0]));
        volume = volume + dot(ptm, ptn);
    }

    volume = volume / 6.0f;
    return volume;
}


/*======================================================================
  = rotate point around normalized axis (using quaternion maths)       =
  ======================================================================*/
void Smoother::rotAroundAxis(float *px, float *py, float *pz, float ax, float ay, float az, float angle)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    float temp;
    float qrot_x, qrot_y, qrot_z, qrot_w;
    float qinv_x, qinv_y, qinv_z, qinv_w;
    float qres_x, qres_y, qres_z, qres_w;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//


    temp = sin(angle / 2.0f);
    qrot_x = ax * temp;
    qrot_y = ay * temp;
    qrot_z = az * temp;
    qrot_w = cos(angle / 2.0f);

    qinv_x = -qrot_x;
    qinv_y = -qrot_y;
    qinv_z = -qrot_z;
    qinv_w = qrot_w;

    qres_x = (qrot_w * (*px))  +  (qrot_y*(*pz) - qrot_z*(*py));
    qres_y = (qrot_w * (*py))  +  (qrot_z*(*px) - qrot_x*(*pz));
    qres_z = (qrot_w * (*pz))  +  (qrot_x*(*py) - qrot_y*(*px));
    qres_w = -( qrot_x*(*px) + qrot_y*(*py) + qrot_z*(*pz) );

    (*px) = (qres_w * qinv_x)  +  (qinv_w * qres_x)  +  (qres_y*qinv_z - qres_z*qinv_y);
    (*py) = (qres_w * qinv_y)  +  (qinv_w * qres_y)  +  (qres_z*qinv_x - qres_x*qinv_z);
    (*pz) = (qres_w * qinv_z)  +  (qinv_w * qres_z)  +  (qres_x*qinv_y - qres_y*qinv_x);
}


/*======================================================================
  = calculate vertex weights for a mesh edge with its adjacent         =
  = triangles (for approx. of second order difference)                 =
  ======================================================================*/
void Smoother::calcWeights(TriMesh::Point& vj, TriMesh::Point& vk, TriMesh::Point& vl1, TriMesh::Point& vl2, float& c_j, float& c_k, float& c_l1, float& c_l2)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    TriMesh::Point vl1_rot;
    TriMesh::Point axis, n_l1, n_l2, tmp_vec;
    float angle, le;
    float u_k, v_k, u_l1, v_l1, u_l2, v_l2;
    float A_j, A_k, A_l1, A_l2;
    TriMesh::Point u_axis, v_axis;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    c_j = c_k = c_l1 = c_l2 = 0.0f;

    // get rotation axis for hinge map
    axis = (vk - vj);
    le = axis.norm();
    if (le != 0.0f)
        axis.normalize();

    // compute rotation angle
    n_l1 = (vl1 - vj) % (vk - vj);          // triangle normal
    if (n_l1.norm() != 0.0f)
        n_l1.normalize();
    n_l2 = (vk - vj) % (vl2 - vj);          // triangle normal
    if (n_l2.norm() != 0.0f)
        n_l2.normalize();

    float cos_angle = (n_l1 | n_l2);
    if (cos_angle > 1.0f)                   // make sure value passed to acos lies in [-1..1]
        cos_angle = 1.0f;
    else if (cos_angle < -1.0f)
        cos_angle = -1.0f;
    angle = acos(cos_angle);

    float orient = (n_l1 % n_l2) | axis;    // clockwise/counter-clockwise test
    if (orient < 0.0f)
        angle = -angle;

    // rotate triangle around edge such that it comes to lie in the
    // same plane as the opposite triangle
    vl1_rot = (vl1 - vj);
    rotAroundAxis(&vl1_rot[0], &vl1_rot[1], &vl1_rot[2], axis[0], axis[1], axis[2], angle);
    vl1_rot = (vl1_rot + vj);

    // project vertices onto triangle plane and get their projected coordinates
    // (the origin is put to the central vertex, the axis are defined as follows)
    u_axis = (vl2 - vj).normalize();
    v_axis = (u_axis % n_l2);

    // project vertices onto plane
    u_l2 = (vl2 - vj).norm();               // lies on u_axis
    v_l2 = 0.0f;
    tmp_vec = (vl1_rot - vj);
    u_l1 = tmp_vec | u_axis;
    v_l1 = tmp_vec | v_axis;
    tmp_vec = (vk - vj);
    u_k = tmp_vec | u_axis;
    v_k = tmp_vec | v_axis;

    // calculate twice the signed triangle areas
    A_l1 = ((u_k - u_l1) * (-v_l1)) - ((v_k - v_l1) * (-u_l1));
    A_l2 = ((-u_l2) * (v_k - v_l2)) - ((-v_l2) * (u_k - u_l2));
    A_k = ((u_l1) * (v_l2)) - ((v_l1) * (u_l2));
    A_j = ((u_l2 - u_k) * (v_l1 - v_k)) - ((v_l2 - v_k) * (u_l1 - u_k));

    // calculate vertex weights
    if (A_l1 != 0.0f)
    {
        A_l1 = 1.0f / A_l1;
        c_l1 = le * A_l1;
    }
    if (A_l2 != 0.0f)
    {
        A_l2 = 1.0f / A_l2;
        c_l2 = le * A_l2;
    }

    if ((A_l1 != 0.0f) && (A_l2 != 0.0f))
    {
        c_k = -le * A_k * A_l1 * A_l2;
        c_j = -le * A_j * A_l1 * A_l2;
    }
}


/*======================================================================
  = biconjugate gradient solver for nonsymm. linear system             =
  =                                                                    =
  = A*x = b   -->   x                                                  =
  =                                                                    =
  = Return values:     >0:   number of iterations used                 =
  =                    -1:   method failed                             =
  =                                                                    =
  = Parameters:                                                        =
  =     [in]    size            problem size                           =
  =     [in]    A               sparse matrix                          =
  =     [in/out]x               in:  the right hand side               =
  =                             out: the approximated solution vector  =
  =     [in]    m               diagonal of inverse preconditioner     =
  =     [in]    max_iter        max. number of iterations              =
  ======================================================================*/
int Smoother::solve_bi_conj_grad(int size, SparseMatrix &A, gsl_vector *x, gsl_vector *m, int max_iter, double eps)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    gsl_vector *r, *r_, *tmp;               // vectors for solver
    gsl_vector *z, *z_, *p, *p_;            // vectors for solver
    double rho1, rho2, beta, alpha, conv_l, conv_r;

    int iter_num;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    rho1 = 0.0;
    rho2 = 0.0;

    r = gsl_vector_alloc(size);
    r_ = gsl_vector_alloc(size);
    tmp = gsl_vector_alloc(size);
    z = gsl_vector_alloc(size);
    z_ = gsl_vector_alloc(size);
    p = gsl_vector_alloc(size);
    p_ = gsl_vector_alloc(size);

    // initial guess for x is right hand side b

    // r = b - A*x
    gsl_vector_memcpy(r, x);
    A.mul_vec(size, gsl_vector_ptr(x, 0), gsl_vector_ptr(tmp, 0));
    gsl_vector_sub(r, tmp);

    gsl_vector_memcpy(r_, r);

    for(iter_num=0; iter_num < max_iter; iter_num++)
    {
        if (m != NULL)
        {
            // solve Mz = r
            // (the precondition matrix M is expected to be diagonal in this case, so solving the
            //  linear system is a vector-vector multiplication)
            gsl_vector_memcpy(z, m);
            gsl_vector_mul(z, r);

            // solve M'z_ = r_
            // (M diagonal)
            gsl_vector_memcpy(z_, m);
            gsl_vector_mul(z_, r_);
        }

        rho2 = rho1;
        gsl_blas_ddot(z, r_, &rho1);

        if (rho1 == 0.0)
        {
            iter_num = -1;                      // method failed
            break;
        }

        if (iter_num == 0)
        {
            gsl_vector_memcpy(p, z);
            gsl_vector_memcpy(p_, z_);
        }
        else
        {
            beta = rho1 / rho2;
            gsl_blas_dscal(beta, p);
            gsl_vector_add(p, z);
            gsl_blas_dscal(beta, p_);
            gsl_vector_add(p_, z_);
        }

        A.mul_vec(size, gsl_vector_ptr(p, 0), gsl_vector_ptr(z, 0));
        A.trans_mul_vec(size, gsl_vector_ptr(p_, 0), gsl_vector_ptr(z_, 0));

        gsl_blas_ddot(p_, z, &alpha);
        alpha = rho1 / alpha;

        // compute right side of convergence criteria
        gsl_vector_memcpy(tmp, x);
        conv_r = eps * gsl_blas_dnrm2(tmp);

        gsl_blas_daxpy(alpha, p, x);

        // compute left side of convergence criteria
        gsl_vector_sub(tmp, x);
        conv_l = gsl_blas_dnrm2(tmp);

        // ||x(i+1) - x(i)||  <=  eps * ||x(i)||
        if (conv_l <= conv_r)
            break;
        
        gsl_blas_daxpy((-alpha), z, r);
        gsl_blas_daxpy((-alpha), z_, r_);
    }

    //A.mul_vec(size, gsl_vector_ptr(x, 0), gsl_vector_ptr(tmp, 0));
    //gsl_vector_sub(tmp, b);
    //printf("\nResidual = %.3e\n", gsl_blas_dnrm2(tmp));

    gsl_vector_free(r);
    gsl_vector_free(r_);
    gsl_vector_free(tmp);
    gsl_vector_free(z);
    gsl_vector_free(z_);
    gsl_vector_free(p);
    gsl_vector_free(p_);

    return iter_num;
}


/*======================================================================
  = smoothing step using implicit euler integration                    =
  ======================================================================*/
void Smoother::step_impl(VPropPoint *pos0, VPropPoint *pos1)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    int size;                       // size for vectors/matrices
    gsl_vector *vec_x = NULL;       // mesh vertex vector (x-coordinates)
    gsl_vector *vec_y = NULL;       // mesh vertex vector (y-coordinates)
    gsl_vector *vec_z = NULL;       // mesh vertex vector (z-coordinates)

    gsl_vector *prec = NULL;        // inverse preconditioner for solving the sparse linear system
    SparseMatrix sp_mat_Smooth;     // "smoothing" matrix: (I - lambda*dt*L) stored in a
                                    // data structure for sparse matrices

    TriMesh::ConstVertexVertexIter cvvIt;
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    const std::string& mode = current_mode_name();
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    //
    // implicit integration:
    //
    // (I - lambda*dt*L) * X(new) = X(old)      (-> solving a linear system)
    //

    assert(pos0 != NULL);
    assert(pos1 != NULL);

    size = mesh_.n_vertices();

    if (size < 1)
        return;

    vec_x = gsl_vector_alloc(size);             // mesh vertex vector
    vec_y = gsl_vector_alloc(size);             // mesh vertex vector
    vec_z = gsl_vector_alloc(size);             // mesh vertex vector
    prec = gsl_vector_alloc(size);

    if(mode == "Umbrella")
    {
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            int index = mesh_.property(vIndex, vIt);            // get vertex index
            TriMesh::Point pt = mesh_.property(*pos0, vIt);     // get vertex coordinates
            
            // insert vertex into mesh vertex vectors
            gsl_vector_set(vec_x, index, pt[0]);
            gsl_vector_set(vec_y, index, pt[1]);
            gsl_vector_set(vec_z, index, pt[2]);

            // set diagonal matrix entry
            sp_mat_Smooth.insert(index, index, (1.0 + int_step_size_));
            gsl_vector_set(prec, index, (1.0 / (1.0 + int_step_size_)));

            // compute valence (number of neighbors)
            int valence = 0;
            for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
            {
                valence++;
            }

            if (valence != 0)
            {
                float factor = 1.0f / (float)valence;
                for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
                {
                    int tmpIndex = mesh_.property(vIndex, cvvIt);

                    // set matrix entry for this neighbor vertex
                    sp_mat_Smooth.insert(index, tmpIndex, (-factor * int_step_size_));
                }
            }
        }
        // nothing to update
    }
    
    else if(mode == "Improved Umbrella")
    {
        float weight, inv_weight;
        
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            int index = mesh_.property(vIndex, vIt);            // get vertex index
            TriMesh::Point pt = mesh_.property(*pos0, vIt);     // get vertex coordinates

            // insert vertex into mesh vertex vectors
            gsl_vector_set(vec_x, index, pt[0]);
            gsl_vector_set(vec_y, index, pt[1]);
            gsl_vector_set(vec_z, index, pt[2]);

            // compute vertex weights over all neighbors
            weight = 0.0f;
            inv_weight = 0.0f;
            for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
            {
                TriMesh::Point v = mesh_.property(*pos0, cvvIt);
                float w = ((pt - v).length());

                if (w != 0.0f)
                {
                    weight = weight + w;
                    inv_weight = inv_weight + (1.0f/w);
                }
            }

            if (weight != 0.0f)
            {
                weight = 2.0f / weight;
                
                // set diagonal matrix entry
                sp_mat_Smooth.insert(index, index, (1.0f + (weight*inv_weight*int_step_size_)));
                gsl_vector_set(prec, index, (1.0 / (1.0f + (weight*inv_weight*int_step_size_))));
                
                for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
                {
                    int tmpIndex = mesh_.property(vIndex, cvvIt);
                    TriMesh::Point v = mesh_.property(*pos0, cvvIt);

                    float w = ((pt - v).length());
                    if (w != 0.0f)
                        w = 1.0f / w;
                    else
                        continue;

                    // set matrix entry for this neighbor vertex
                    sp_mat_Smooth.insert(index, tmpIndex, (-weight * w * int_step_size_));
                }
            }
        }
        // nothing to update
    }
    else if(mode == "Curvature Flow")
    {
        typedef std::pair<float, float> PairFloat;

        int valence, i;
        float weight, sum;
        PairFloat cot;
        std::vector<PairFloat> cots;

        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            int index = mesh_.property(vIndex, vIt);            // get vertex index
            TriMesh::Point center = mesh_.property(*pos0, vIt); // get vertex coordinates

            // insert vertex into mesh vertex vectors
            gsl_vector_set(vec_x, index, center[0]);
            gsl_vector_set(vec_y, index, center[1]);
            gsl_vector_set(vec_z, index, center[2]);

            // set diagonal matrix entry
            sp_mat_Smooth.insert(index, index, (1.0 + int_step_size_));
            gsl_vector_set(prec, index, (1.0 / (1.0 + int_step_size_)));

            sum = 0.0f;

            // compute cotangens
            for(cvvIt = mesh_.cvv_iter(vIt), valence = 0; cvvIt; ++valence)
            {
                TriMesh::Point& v0 = mesh_.property(*pos0, cvvIt);
                TriMesh::Point& v1 = mesh_.property(*pos0, ++cvvIt);
    
                try {
                    cot.first  = cotangent(v1 - v0, center - v0);
                    cot.second = cotangent(v0 - v1, center - v1);
                } catch(int) {
                    // area of triangle is 0
                    cot.first = cot.second = 0;
                }
    
                cots.push_back(cot);
                sum = sum + cot.first + cot.second;
            }

            for(cvvIt = mesh_.cvv_iter(vIt), i = valence - 1; cvvIt; ++cvvIt)
            {
                int tmpIndex = mesh_.property(vIndex, cvvIt);

                weight = cots[i].first + cots[++i %= valence].second;

                // set matrix entry for this neighbor vertex
                if (sum != 0.0f)
                    sp_mat_Smooth.insert(index, tmpIndex, ((-weight / sum) * int_step_size_));
            }

            cots.clear();
        }
    }

    if(mode == "2nd Order Difference")
    {
        int valence, i, flaps = 0;
        float sum;
        std::vector<TriMesh::VertexHandle> one_ring;
        std::vector<float> weights;

        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            int index = mesh_.property(vIndex, vIt);            // get vertex index
            int tmpIndex;
            TriMesh::Point pt = mesh_.property(*pos0, vIt);     // get vertex coordinates
            
            // insert vertex into mesh vertex vectors
            gsl_vector_set(vec_x, index, pt[0]);
            gsl_vector_set(vec_y, index, pt[1]);
            gsl_vector_set(vec_z, index, pt[2]);

            // set diagonal matrix entry
            sp_mat_Smooth.insert(index, index, (1.0 + int_step_size_));
            gsl_vector_set(prec, index, (1.0 / (1.0 + int_step_size_)));

            // compute valence and store references to one ring neighbors
            valence = 0;
            for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
            {
                valence++;
                one_ring.push_back(cvvIt.handle());

                weights.push_back(0.0f);
            }

            sum = 0.0f;
            for(i=0; i < valence; i++)
            {
                TriMesh::Point vl1, vl2, vk;
                float c_j, c_k, c_l1, c_l2;

                vk = mesh_.property(*pos0, one_ring[i]);
                vl1 = mesh_.property(*pos0, one_ring[(i + (valence-1)) % valence]);
                vl2 = mesh_.property(*pos0, one_ring[(i+1) % valence]);

                calcWeights(pt, vk, vl1, vl2, c_j, c_k, c_l1, c_l2);

                weights[i] += (c_j * c_k);          // update weight for this vertex
                weights[(i + (valence-1)) % valence] += (c_j * c_l1);   // update weight for this vertex
                weights[(i+1) % valence] += (c_j * c_l2);   // update weight for this vertex

                sum = sum + (c_j * c_j);
            }

            // process vertices of "flap" triangles
            /*for(i=0; i < valence; i++)
            {
                TriMesh::Point vl2, vj, vk;
                float c_j, c_k, c_l1, c_l2;
                TriMesh::VertexHandle vh_k, vh_j;
                TriMesh::HalfedgeHandle hh;

                vh_k = one_ring[i];
                vk = mesh_.property(*pos0, vh_k);
                vh_j = one_ring[(i+1) % valence];
                vj = mesh_.property(*pos0, vh_j);

                // find vertex of "flap" triangle
                TriMesh::ConstVertexOHalfedgeIter cvohIt;
                for(cvohIt = mesh_.cvoh_iter(vh_k); cvohIt; ++cvohIt)
                {
                    if (mesh_.to_vertex_handle(cvohIt) == vh_j)
                        break;
                }
                if (cvohIt == NULL)
                    continue;

                hh = mesh_.cw_rotated_halfedge_handle(cvohIt);
                if (mesh_.to_vertex_handle(hh) == vIt.handle())
                    hh = mesh_.ccw_rotated_halfedge_handle(cvohIt);
                
                vl2 = mesh_.property(*pos0, mesh_.to_vertex_handle(hh));

                calcWeights(vj, vk, pt, vl2, c_j, c_k, c_l1, c_l2);

                flaps++;
                one_ring.push_back(mesh_.to_vertex_handle(hh));
                weights.push_back((c_l1 * c_l2));           // set weight for this vertex

                weights[i] += (c_l1 * c_k);                 // update weight for this vertex
                weights[(i+1) % valence] += (c_l1 * c_j);   // update weight for this vertex

                sum = sum + (c_l1 * c_l1);
            }*/
            
            if (sum != 0.0f)
            {
                sum = -1.0f / sum;
                
                for(i=0; i < (valence+flaps); i++)
                {
                    tmpIndex = mesh_.property(vIndex, one_ring[i]); // get vertex index

                    // set matrix entry for this neighbor vertex
                    sp_mat_Smooth.insert(index, tmpIndex, (-sum * int_step_size_ * weights[i]));
                }
            }

            one_ring.clear();
            weights.clear();
        }
    }


    // solve linear system of equations (with multiple right-hand sides)
    // to get the smoothed vertices
    int info;
    info = solve_bi_conj_grad(size, sp_mat_Smooth, vec_x, prec);
    info = info + solve_bi_conj_grad(size, sp_mat_Smooth, vec_y, prec);
    info = info + solve_bi_conj_grad(size, sp_mat_Smooth, vec_z, prec);
    printf("Linear solver: %d iteration(s) (mean value)\n", (info / 3));
    
    // copy the smoothed vertices to the mesh
    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
    {
        int index = mesh_.property(vIndex, vIt);
        TriMesh::Point tmpPt(gsl_vector_get(vec_x, index), gsl_vector_get(vec_y, index), gsl_vector_get(vec_z, index));

        TriMesh::Point& pos = mesh_.property(*pos1, vIt);
        pos = tmpPt;
    }

    // free space
    gsl_vector_free(vec_x);
    gsl_vector_free(vec_y);
    gsl_vector_free(vec_z);
}

void Smoother::step_expl(VPropPoint *pos0, VPropPoint *pos1)
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    TriMesh::ConstVertexVertexIter cvvIt;
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    const std::string& mode = current_mode_name();
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//


    assert(pos0 != NULL);
    assert(pos1 != NULL);

    if(mode == "Umbrella")
    {
        int valence;
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {

            TriMesh::Point& pos = mesh_.property(*pos1, vIt).vectorize(0);
            for(cvvIt = mesh_.cvv_iter(vIt), valence = 0; cvvIt; ++cvvIt, ++valence)
                pos += mesh_.property(*pos0, cvvIt);
            if(valence > 0)
            {
                pos *= (int_step_size_ / valence);
                pos += ((1.0f - int_step_size_) * mesh_.property(*pos0, vIt));
            }
            else pos = mesh_.property(*pos0, vIt);
        }
        // nothing to update
    } 
    
    else if(mode == "Improved Umbrella")
    {
        float sum, weight, inv_weight;
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            TriMesh::Point& center = mesh_.property(*pos0, vIt);
            TriMesh::Point& pos = mesh_.property(*pos1, vIt).vectorize(0);
            for(cvvIt = mesh_.cvv_iter(vIt), inv_weight = 0, sum = 0; cvvIt; ++cvvIt)
            {
                TriMesh::Point& v = mesh_.property(*pos0, cvvIt);
                if((weight = (center - v).length()) == 0)
                { 
                    sum = 0; 
                    break;
                } else {
                    sum += weight;
                    weight = 1.0 / weight;
                    inv_weight += weight;
                    pos += v * weight;
                }
            }
            if(sum > 0)
            {
                sum = 2.0 / sum;
                pos *= (int_step_size_ * sum);
                pos += ((1.0f - (int_step_size_ * sum * inv_weight)) * center);
            }
            else pos = center;
        }
    }

    else if(mode == "Curvature Flow")
    {
        typedef std::pair<float, float> PairFloat;

        int valence, i;
        float weight, sum;
        PairFloat cot;
        std::vector<PairFloat> cots;

        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            TriMesh::Point& center = mesh_.property(*pos0, vIt);
            TriMesh::Point& pos = mesh_.property(*pos1, vIt).vectorize(0);

            // compute cotangens
            for(cvvIt = mesh_.cvv_iter(vIt), valence = 0; cvvIt; ++valence)
            {
                TriMesh::Point& v0 = mesh_.property(*pos0, cvvIt);
                TriMesh::Point& v1 = mesh_.property(*pos0, ++cvvIt);
    
                try {
                    cot.first  = cotangent(v1 - v0, center - v0);
                    cot.second = cotangent(v0 - v1, center - v1);
                } catch(int) {
                    // area of triangle is 0
                    cot.first = cot.second = 0;
                }
    
                cots.push_back(cot);
            }

            for(cvvIt = mesh_.cvv_iter(vIt), sum = 0, i = valence - 1; cvvIt; ++cvvIt)
            {
                weight  = cots[i             ].first;
                weight += cots[++i %= valence].second;
                pos += mesh_.property(*pos0, cvvIt) * weight;
                sum += weight;
            }
            if(sum > 0)
            {
                pos *= (int_step_size_ / sum);
                pos += ((1.0f - int_step_size_) * center);
            }
            else pos = center;

            cots.clear();
        }
    }

    else if (mode == "2nd Order Difference")
    {
        int valence, i;
        float sum;
        std::vector<TriMesh::VertexHandle> one_ring;

        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        {
            TriMesh::Point& center = mesh_.property(*pos0, vIt);
            TriMesh::Point& pos = mesh_.property(*pos1, vIt).vectorize(0);

            // compute valence and store references to one ring neighbors
            valence = 0;
            for(cvvIt = mesh_.cvv_iter(vIt); cvvIt; ++cvvIt)
            {
                valence++;
                one_ring.push_back(cvvIt.handle());
            }

            sum = 0.0f;
            for(i=0; i < valence; i++)
            {
                TriMesh::Point vl1, vl2, vk;
                TriMesh::Point de;
                float c_j, c_k, c_l1, c_l2;

                vk = mesh_.property(*pos0, one_ring[i]);
                vl1 = mesh_.property(*pos0, one_ring[(i + (valence-1)) % valence]);
                vl2 = mesh_.property(*pos0, one_ring[(i+1) % valence]);

                calcWeights(center, vk, vl1, vl2, c_j, c_k, c_l1, c_l2);

                // calculate vertex weights and apply update procedure
                de = c_j * ((c_l1*vl1) + (c_l2*vl2) + (c_k*vk));

                pos = pos + de;
                sum = sum + (c_j * c_j);
            }

            /*// process vertices of "flap" triangles
            for(i=0; i < valence; i++)
            {
                TriMesh::Point vl2, vj, vk;
                TriMesh::Point de;
                float c_j, c_k, c_l1, c_l2;
                TriMesh::VertexHandle vh_k, vh_j;
                TriMesh::HalfedgeHandle hh;

                vh_k = one_ring[i];
                vk = mesh_.property(*pos0, vh_k);
                vh_j = one_ring[(i+1) % valence];
                vj = mesh_.property(*pos0, vh_j);

                // find vertex of "flap" triangle
                TriMesh::ConstVertexOHalfedgeIter cvohIt;
                for(cvohIt = mesh_.cvoh_iter(vh_k); cvohIt; ++cvohIt)
                {
                    if (mesh_.to_vertex_handle(cvohIt) == vh_j)
                        break;
                }
                if (cvohIt == NULL)
                    continue;

                hh = mesh_.cw_rotated_halfedge_handle(cvohIt);
                if (mesh_.to_vertex_handle(hh) == vIt.handle())
                    hh = mesh_.ccw_rotated_halfedge_handle(cvohIt);
                
                vl2 = mesh_.property(*pos0, mesh_.to_vertex_handle(hh));

                calcWeights(vj, vk, center, vl2, c_j, c_k, c_l1, c_l2);

                // calculate vertex weights and apply update procedure
                de = c_l1 * ((c_j*vj) + (c_l2*vl2) + (c_k*vk));

                pos = pos + de;
                sum = sum + (c_l1 * c_l1);
            }*/
            
            if (sum != 0.0f)
            {
                pos *= (-int_step_size_ / sum);
                pos += ((1.0f - int_step_size_) * center);
            }
            else
                pos = center;

            one_ring.clear();
        }
    }
}

void Smoother::step()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    VPropPoint pos1, pos0 = posVec.back();
    mesh_.add_property(pos1);
    posVec.push_back(pos1);

    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
    float vol0, vol1;
    float factor;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//


    if (int_mode_ == eIntExpl)
        step_expl(&pos0, &pos1);    // explicit integration
    else
        step_impl(&pos0, &pos1);    // implicit integration

    if (vol_pres_)
    {   // volume preservation enabled
        vol0 = mesh_volume(&pos0);
        vol1 = mesh_volume(&pos1);

        if ((vol0 != 0.0f) && (vol1 != 0.0f))
        {
            // apply volume correction
            factor = pow((vol0 / vol1), (1.0f/3.0f));
            for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
            {
                TriMesh::Point &pt = mesh_.property(pos1, vIt);
                pt = factor * pt;
            }
        }
    }
}

void Smoother::apply(int _iter, float _alpha)
{
    TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

    if(_iter == iter_count_)
        --_iter, _alpha = 1;

    VPropPoint pos1 = posVec[_iter + 1],
               pos0 = posVec[_iter];

    // set distances
    if(cur_iter_ != _iter)
    {
        for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
            mesh_.property(distances, vIt) = 
                (mesh_.property(pos1, vIt) - mesh_.property(pos0, vIt)).length();
        cur_iter_ = _iter;
    }

    for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
        mesh_.set_point(vIt, lerp(mesh_.property(pos0, vIt), mesh_.property(pos1, vIt), _alpha));

    // all properties need to be updated
    mesh_.update_later();
}


/*======================================================================
  = run algorithm (overridden base class function to do algorithm      =
  = specific tasks)                                                    =
  ======================================================================*/
void Smoother::run()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    QTime time;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//

    // check and set step size from input
    if (!set_int_step_size())
    {
        QString s;
        s.sprintf("Please enter a valid step size in\n[%.3e, %.3e]", min_int_step_, max_int_step_);
        QMessageBox::information(NULL, "Smoother information", s);
        return;
    }

    time.start();                   // do time measurement

    Algorithm::run();               // call base class function

    printf("Smoothing computation time (%d vertices, %d iteration(s)):\t%.2e ms\n", mesh_.n_vertices(), iter_count_, (double)time.elapsed());     // show cumputation time
}
