#pragma once

#include "algorithm.h"
#include "smootherwidget.h"

// representation of a sparse matrix
#include "sparsematrix.h"

// stl
#include <limits>

// gsl (solving linear equation systems etc.)
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

// qt
#include <qdatetime.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qspinbox.h>
#include <qslider.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qbuttongroup.h>


class Smoother : public Algorithm
{
    Q_OBJECT

	typedef OpenMesh::VPropHandleT<TriMesh::Point> VPropPoint;
	typedef OpenMesh::VPropHandleT<float>          VPropFloat;
    typedef OpenMesh::VPropHandleT<int>            VPropInt;
	typedef OpenMesh::PropertyT<TriMesh::Point>    Points;

    // integration method
    enum enumIntMode
    {
        eIntExpl,           // explicit integration
        eIntImpl,           // implicit integration
    };

public:
	Smoother(TriMesh& _mesh, QObject* _parent = 0, const char* _name = 0);
	virtual ~Smoother();

public:
	virtual const std::string& draw_mode_name();

public slots:
    void run();                         // run the algorithm
    void set_int_mode(int id);          // set the integration mode
    void set_vol_pres(void);            // enable/disable volume preservation

protected:
	virtual void init();
	virtual void step();
	virtual void apply(int _iter, float _alpha);

private:
    bool set_int_step_size(void);                       // set the integration step size

    float mesh_volume(VPropPoint *pos);

    void step_expl(VPropPoint *pos0, VPropPoint *pos1); // step with explicit integration
    void step_impl(VPropPoint *pos0, VPropPoint *pos1); // step with implicit integration

	float cotangent(const TriMesh::Point& v0, const TriMesh::Point& v1)
	{
		float f = (v0 % v1).length();
		if(f == 0) throw 0;
		return (v0 | v1) / f;
	}

    inline int solve_bi_conj_grad(int size, SparseMatrix &A, gsl_vector *x, gsl_vector *m=NULL, int max_iter=100, 
                                  double eps=std::numeric_limits<double>::epsilon());

    inline void rotAroundAxis(float *px, float *py, float *pz, float ax, float ay, float az, float angle);
    inline void calcWeights(TriMesh::Point& vj, TriMesh::Point& vk, TriMesh::Point& vl1, TriMesh::Point& vl2, float& c_j, float& c_k, float& c_l1, float& c_l2);

	static std::string draw_mode_;
    enumIntMode int_mode_;
    float min_int_step_, max_int_step_;
    float int_step_size_;               // integration step size
    bool vol_pres_;

	std::vector<VPropPoint> posVec;
	VPropFloat distances;
    VPropInt vIndex;                    // vertex index

	int cur_iter_;

    SmootherWidget *smootherWdg_;       // smoother widget
};
