#include "subdivider.h"

// openmesh
#include <OpenMesh\Tools\Subdivider\Uniform\LoopT.hh>
#include <OpenMesh\Tools\Subdivider\Uniform\Sqrt3T.hh>

#include "ButterflyT.h"

std::string Subdivider::draw_mode_ = "CV:Subdivision Vertices";

template <typename T>
const T Subdivider::Value<T>::inv_ = std::numeric_limits<T>::max();

Subdivider::Subdivider(TriMesh& _mesh, QObject* _parent, const char* _name)
: Algorithm(_mesh, _parent, _name), subdivider_(0), cur_iter_(-1), old_iter_count_(0)
{
    mesh_.add_property(iterValue);
	mesh_.add_property(iterFloat, "Subdivision Vertices");

    subdividerWdg_ = NULL;

    if (_parent != NULL)
    {
        // create and add subdivider widget
        subdividerWdg_ = new SubdividerWidget(scrollView_->viewport(), "SubdividerWidget");
        scrollView_->addChild(subdividerWdg_);
        scrollView_->viewport()->setPaletteBackgroundColor(subdividerWdg_->paletteBackgroundColor());
        
        // choose settings for tool bar
        toolBar_->setCaption("Subdivider");
        // check whether widget fits into dock area and adjust size
        QSize size = toolBar_->sizeHint();
        QDockArea* area = toolBar_->area();
        if (area != NULL)   // toolbar docked
        {
            int width;
            if ((area->height()-20) > subdividerWdg_->minimumHeight())
                width = (size.width() / 2) + subdividerWdg_->minimumWidth();
            else
                width = size.width() + scrollView_->sizeHint().width();
            toolBar_->setFixedExtentWidth(width);
        }
        
        // set control elements
        set_modeBox(subdividerWdg_->modeComboBox());
        set_slider(subdividerWdg_->frameSlider());
        set_playButton(subdividerWdg_->playPushButton());
        
        connect(subdividerWdg_->modeComboBox(), SIGNAL(activated(int)), SLOT(set_mode(int)));
	    connect(subdividerWdg_->iterCountSpinBox(), SIGNAL(valueChanged(int)), SLOT(set_iter_count(int)));
        connect(subdividerWdg_->runPushButton(), SIGNAL(clicked()), SLOT(run()));
	    connect(subdividerWdg_->playPushButton(), SIGNAL(pressed()), SLOT(animate()));
	    connect(subdividerWdg_->frameSlider(), SIGNAL(valueChanged(int)), SLOT(set_frame(int)));
        
        add_mode("Loop");
	    add_mode("Sqrt3");
        add_mode("Butterfly");
        add_mode("Modified Butterfly");

        // set first subdivider mode
        if (n_modes_ >= 1)
            set_mode(1);
    }
}

Subdivider::~Subdivider()
{
	apply_topology();

	std::vector<VPropPoint>::iterator it, end = posVec.end();
	for(it = posVec.begin(); it != end; ++it)
		mesh_.remove_property(*it);

	mesh_.remove_property(iterValue);
	mesh_.remove_property(iterFloat);

    // subdivider widget will be deleted automatically by parent object
}

const std::string& Subdivider::draw_mode_name()
{
	return draw_mode_;
}

void Subdivider::init()
{
	// apply current subdivision level
	apply_topology();

    // store old iteration counter
    old_iter_count_ = iter_count_;
	
	// delete all pos
	for( ; !posVec.empty(); posVec.pop_back())
		mesh_.remove_property(posVec.back());

	// add initial position
	VPropPoint pos0;
	mesh_.add_property(pos0);
	posVec.push_back(pos0);

	TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
	{
		mesh_.property(pos0, vIt) = mesh_.point(vIt);
		mesh_.property(iterValue, vIt).set_value(0);
        mesh_.property(iterFloat, vIt) = 0.0f;
	}

	if(subdivider_)
	{
		delete subdivider_;
	}


	typedef OpenMesh::Subdivider::Uniform::LoopT<TriMesh>           LoopT;
	typedef OpenMesh::Subdivider::Uniform::Sqrt3T<TriMesh>          Sqrt3T;
    typedef ButterflyT<TriMesh>      ButterflyT;

	const std::string& mode = current_mode_name();

	if(mode == "Loop")
		subdivider_ = new LoopT();
	else if(mode == "Sqrt3")
		subdivider_ = new Sqrt3T();
    else if(mode == "Butterfly")
        subdivider_ = new ButterflyT(ButterflyT::eOriginal);
    else if(mode == "Modified Butterfly")
        subdivider_ = new ButterflyT(ButterflyT::eModified);

	subdivider_->attach(mesh_);
	cur_iter_ = 0;
}

void Subdivider::step()
{
	++cur_iter_;

	std::cout << "Subdividing from " << mesh_.n_vertices() << " to ";
	subdivider_->operator()(1);
	std::cout << mesh_.n_vertices() << " vertices.\n";

	VPropPoint pos1;
	mesh_.add_property(pos1);
	posVec.push_back(pos1);

	int valence;
	std::vector<VPropPoint>::const_iterator pIt, pEnd = posVec.end()-1;
	TriMesh::ConstVertexVertexIter cvvIt;
	TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

	// iterate vertices
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
	{
		// if old frame position of new vertex has not been set
		if(!mesh_.property(iterValue, vIt).valid())
		{
			// iterate all old frames
            // (new points: the positions for all previous steps have to be set)
			for(pIt = posVec.begin(); pIt != pEnd; ++pIt)
			{
				TriMesh::Point& pos = mesh_.property(*pIt, vIt).vectorize(0);
				// iterate all neighboring vertices
				for(cvvIt = mesh_.cvv_iter(vIt), valence = 0; cvvIt; ++cvvIt)
				{
					// if not new vertex
					if(mesh_.property(iterValue, cvvIt).value() < cur_iter_)
					{
						++valence;
						pos += mesh_.property(*pIt, cvvIt);
					}
				}
                pos *= 1.0f / valence;
			}
			mesh_.property(iterValue, vIt).set_value(cur_iter_);
            mesh_.property(iterFloat, vIt) = cur_iter_;
		}
		mesh_.property(pos1, vIt) = mesh_.point(vIt);
	}

}

void Subdivider::postprocess()
{
	cur_iter_ = -1;
}

void Subdivider::apply(int _iter, float _alpha)
{
	TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

	if(_iter == iter_count_)
		--_iter, _alpha = 1;

	VPropPoint pos1 = posVec[_iter + 1],
			   pos0 = posVec[_iter];

	// set iteration
	if(cur_iter_ != (_iter += (_alpha > 0.0f)))
	{
		cur_iter_ = _iter;
	}

	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
		mesh_.set_point(vIt, lerp(mesh_.property(pos0, vIt), mesh_.property(pos1, vIt), _alpha));

	// all properties need to be updated
	mesh_.update_later();
}

void Subdivider::apply_topology()
{
	if(cur_iter_ < 0 || cur_iter_ >= old_iter_count_)
		return;

	// holding the edges to collapse
	typedef std::vector<TriMesh::HalfedgeHandle> HhVec;
	HhVec hhVec;

	TriMesh::ConstVertexOHalfedgeIter cvohIt;
	TriMesh::VertexIter vIt, vEnd = mesh_.vertices_end();

	// iterate vertices
	for(vIt = mesh_.vertices_begin(); vIt != vEnd; ++vIt)
	{
		// if vertex was added in a later iteration
		if(mesh_.property(iterValue, vIt).value() > cur_iter_)
		{
			// find a neighbour that already exists in the current iteration
			for(cvohIt = mesh_.cvoh_iter(vIt); 
				(cvohIt && !(mesh_.property(iterValue, mesh_.to_vertex_handle(cvohIt)).value() <= cur_iter_));
				++cvohIt);
            if (cvohIt != NULL)
			    hhVec.push_back(cvohIt.handle());
		}
	}

	// now collapse the edges
	for(HhVec::const_iterator hhIt = hhVec.begin(), hhEnd = hhVec.end(); hhIt != hhEnd; ++hhIt)
		mesh_.collapse(*hhIt);


	mesh_.garbage_collection();

#if defined(_DEBUG) || defined(DEBUG)
      // Now we have an consistent mesh!
      assert( OpenMesh::Utils::MeshCheckerT<TriMesh>(mesh_).check() );
#endif

	mesh_.update_later();
}