#pragma once

// stl
#include <limits>
#include <vector>
// gsl
#include <gsl/gsl_linalg.h>
// boost
#include <boost/bind.hpp>

#include "trimesh.h"

std::string TriMesh::prop_cv_mode_ = "Vertex Property";

// constant color
class TriMesh::ConstColorer : public TriMesh::Colorer
{
public:
	ConstColorer(TriMesh& _mesh, const Color& _color) : Colorer(_mesh), color_(_color) {}
	Color operator()(const VertexHandle& _vh) const { return color_; }
private:
	const Color& color_;
};

class TriMesh::NormalColorer : public TriMesh::Colorer
{
public:
	NormalColorer(TriMesh& _mesh) : Colorer(_mesh) {}
	Color operator()(const VertexHandle& _vh) const
	{
		const Point& n = mesh_.normal(_vh);
		return Color(fabs(n[0])*255, fabs(n[1])*255, fabs(n[2])*255);
	}
};

// visualize a scalar function with hue
class TriMesh::HueColorer : public TriMesh::Colorer
{
	typedef float (Vertex::*VFunPtr)() const;

	Color hue2rgb(float hue) const
	{
		hue *= 6;
		int   i = static_cast<int>(hue);
		float f = hue - static_cast<float>(i);
		switch (i)
		{
			case 0: return Color( 255, 255*f, 0 );
			case 1: return Color( 255*(1.0f - f), 255, 0 );
			case 2: return Color( 0, 255, 255*f );
			case 3: return Color( 0, 255*(1.0f - f), 255 );
			case 4: return Color( 255*f, 0, 255 );
			case 5: return Color( 255, 0, 255*(1.0f - f) );
		}
		return Color(0, 0, 0);
	}

	template<typename UnaryFunction>
	void init(const UnaryFunction& _fun)
	{
		mesh_.add_property(funVals_);

		// store vertex values in property
		ConstVertexIter cvIt, cvEnd = mesh_.vertices_end();
		for(cvIt = mesh_.vertices_begin(); cvIt != cvEnd; ++cvIt)
			mesh_.property(funVals_, cvIt.handle()) = _fun(cvIt.handle());
		init(funVals_);
	}

	void init(const VPropHandleFloat& _vPropHandle)
	{
		// qunatile heaps
		unsigned int n = std::max<unsigned int>(1, mesh_.n_vertices() * 0.05f);
		std::vector<float> 
			min(n,  std::numeric_limits<float>::max()),
			max(n, -std::numeric_limits<float>::max());

		ConstVertexIter cvIt, cvEnd = mesh_.vertices_end();
		for(cvIt = mesh_.vertices_begin(); cvIt != cvEnd; ++cvIt)
		{
			min.back() = max.back() = mesh_.property(_vPropHandle, cvIt.handle());
			if(min.front() > min.back())
			{
				push_heap(min.begin(), min.end(), std::less<float>());
				pop_heap (min.begin(), min.end(), std::less<float>());
			}
			if(max.front() < min.back())
			{
				push_heap(max.begin(), max.end(), std::greater<float>());
				pop_heap (max.begin(), max.end(), std::greater<float>());
			}
		}
		hScale_ = max.front() > (hMin_ = min.front()) ? 1.0f / (max.front() - hMin_) : 1.0f;
	}

public:

	HueColorer(TriMesh& _mesh, VFunPtr _vFunPtr) 
		: Colorer(_mesh), myProp_(true)
	{
		const Vertex& (TriMesh::*access)(VertexHandle) const = &TriMesh::vertex;
		init(boost::bind(_vFunPtr, boost::bind(access, boost::cref(mesh_), _1)));
	}

	HueColorer(TriMesh& _mesh, const VPropHandleFloat& _vPropHandle) 
		: Colorer(_mesh), funVals_(_vPropHandle), myProp_(false)
	{
		init(_vPropHandle);
	}

	~HueColorer()
	{
		if(myProp_)
			mesh_.remove_property(funVals_);
	}

	Color operator()(const VertexHandle& _vh) const
	{
		float hue = hScale_ * (mesh_.property(funVals_, _vh) - hMin_);
		return hue2rgb((1.0f - std::max(0.0f, std::min(1.0f, hue))) * (2.0 / 3));
	}

protected:
	VPropHandleFloat funVals_;
	float hScale_, hMin_;
	bool myProp_;
};

TriMesh::TriMesh()
: update_bits_(0)
{
}


/*======================================================================
  = initialize texture coordinates by using spherical texture mapping  =
  ======================================================================*/
void TriMesh::init_vertex_texcoords()
{
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    int n;
    TexCoord tex_coord;                     // texture coordinates
    Point mass_point;

    float max_val = -std::numeric_limits<float>::max();
    float min_val = std::numeric_limits<float>::max();
    float range;
    
    VertexIter vIt, vEnd = vertices_end();
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -//
    

	n = n_vertices();
    if (n <= 0)
        return;

    // compute mass point
    mass_point.vectorize(0);
    for(vIt = vertices_begin(); vIt != vEnd; ++vIt)
	{
        const Point& pt = point(vIt);

        if (pt[1] > max_val)
            max_val = pt[1];
        if (pt[1] < min_val)
            min_val = pt[1];

        mass_point += pt;
    }
    mass_point *= (1.0 / n);
    range = max_val - min_val;

    if (range == 0.)
        return;
    
    for(vIt = vertices_begin(); vIt != vEnd; ++vIt)
	{
		const VertexHandle& vh = vIt.handle();
        Point pt = point(vh);
        pt = pt - mass_point;

        /*tex_coord[0] = atan2(pt[0], pt[2]);
        tex_coord[0] = tex_coord[0] + M_PI;         // map into interval [0..2PI]
        tex_coord[0] = tex_coord[0] / (2.*M_PI);    // normalize*/

        tex_coord[0] = atan2(pt[0], fabs(pt[2]));
        tex_coord[0] = tex_coord[0] + M_PI_2;       // map into interval [0..PI]
        tex_coord[0] = tex_coord[0] / M_PI;         // normalize

        // adjust extremal cases to prevent from texturing artifacts
        if (tex_coord[0] >= 1.0f)
            tex_coord[0] = .999f;
        if (tex_coord[0] <= 0.0f)
            tex_coord[0] = 0.001f;

        /*tex_coord[1] = atan2(pt[1], sqrt((pt[0]*pt[0])+(pt[2]*pt[2])));
        tex_coord[1] = tex_coord[1] + M_PI_2;       // map into interval [0..PI]
        tex_coord[1] = tex_coord[1] / M_PI;         // normalize*/
        
        tex_coord[1] = (pt[1] + mass_point[1] - min_val) / range;

        // adjust extremal cases to prevent from texturing artifacts
        if (tex_coord[1] >= 1.0f)
            tex_coord[1] = .999f;
        if (tex_coord[1] <= 0.0f)
            tex_coord[1] = 0.001f;

        set_texcoord(vh, tex_coord);
    }
}


void TriMesh::update_vertex_curvatures()
{
	gsl_vector* tau = gsl_vector_alloc(3);
	gsl_vector* hes = gsl_vector_alloc(3);

	VertexIter vIt, vEnd = vertices_end();
	for(vIt = vertices_begin(); vIt != vEnd; ++vIt)
	{
		const VertexHandle& vh = vIt.handle();
		const Point& p = point(vh);
		const Point& n = normal(vh);

		Point tang = (fabsf(n[0]) > -0.5f ? Point(n[2], 0, -n[0]) : Point(0, n[2], -n[1])).normalize();
		Point bitang  = n % tang;

		// count valency
		int valency = 0;
		ConstVertexVertexIter cvvIt;
		for(cvvIt = cvv_iter(vh); cvvIt; ++cvvIt) 
			++valency;

		if(valency < 3) 
		{
			vIt->set_first_curvature(0);
			vIt->set_second_curvature(0);
			vIt->set_first_tangent(Point(0.0f, 0.0f, 0.0f));
			continue;
		}

		gsl_matrix* mat = gsl_matrix_alloc(valency, 3);
		gsl_vector* rhs = gsl_vector_alloc(valency);
		gsl_vector* res = gsl_vector_alloc(valency);

		// setup matrix
		int i = 0;
		for(cvvIt = cvv_iter(vh); cvvIt; ++cvvIt, ++i)
		{
			Point dp = point(cvvIt.handle()) - p;

			float u = tang   | dp; 
			float v = bitang | dp; 

			gsl_matrix_set(mat, i, 0,   u*u);
			gsl_matrix_set(mat, i, 1, 2*u*v);
			gsl_matrix_set(mat, i, 2,   v*v);

			gsl_vector_set(rhs, i, n | dp); 
		}

		gsl_linalg_QR_decomp(mat, tau);
		gsl_linalg_QR_lssolve(mat, tau, rhs, hes, res);

		double a = gsl_vector_get(hes, 0);
		double b = gsl_vector_get(hes, 1);
		double c = gsl_vector_get(hes, 2);

		double s = sqrtf((a-c)*(a-c) + 4.0f * b * b);

		vIt->set_first_curvature ((a + c + s) / 2);
		vIt->set_second_curvature((a + c - s) / 2);

		vIt->set_first_tangent((bitang * (b * 2) + tang * (a - c + s)).normalize());

		gsl_matrix_free(mat);
		gsl_vector_free(rhs);
		gsl_vector_free(res);
	}

	gsl_vector_free(tau);
	gsl_vector_free(hes);
}

void TriMesh::update_vertex_colors()
{
	if(cv_mode_ == "Constant")
		update_vertex_colors(ConstColorer(*this, Color(255, 255, 255)));

	else if (cv_mode_ == "Normal")
		update_vertex_colors(NormalColorer(*this));

	else if (cv_mode_ == "1st Principle Curvature")
		update_vertex_colors(HueColorer(*this, Vertex::first_curvature));

	else if (cv_mode_ == "2nd Principle Curvature")
		update_vertex_colors(HueColorer(*this, Vertex::second_curvature));

	else if (cv_mode_ == "Mean Curvature")
		update_vertex_colors(HueColorer(*this, Vertex::mean_curvature));

	else if (cv_mode_ == "Gaussian Curvature")
		update_vertex_colors(HueColorer(*this, Vertex::gaussian_curvature));

	else if (cv_mode_ == "Shape Index")
		update_vertex_colors(HueColorer(*this, Vertex::shape_index));

	else if (cv_mode_ == prop_cv_mode_ && prop_cv_handle_.is_valid())
		update_vertex_colors(HueColorer(*this, prop_cv_handle_));

	else
		std::cerr << "TriMesh: vertex_color_mode not supported.\n";
}

void TriMesh::update_later(unsigned int _bits)
{
	update_bits_ |= _bits;
}

void TriMesh::update_properties(unsigned int _bits, bool _force)
{
	if(_bits & ColorBit)
	{
		if( cv_mode_ == "Normal")
			_bits |= NormalBit;

		if( cv_mode_ == "1st Principle Curvature" ||
			cv_mode_ == "2nd Principle Curvature" ||
			cv_mode_ == "Mean Curvature" ||
			cv_mode_ == "Gaussian Curvature" ||
			cv_mode_ == "Shape Index")
			_bits |= CurvatureBit;
	}

	if(!_force)
		_bits &= update_bits_;

	if(_bits & NormalBit)
	{
		update_normals();
		update_bits_ &= ~NormalBit;
	}
	if(_bits & CurvatureBit)
	{
		update_vertex_curvatures();
		update_bits_ &= ~CurvatureBit;
	}
	if(_bits & ColorBit)
	{
		update_vertex_colors();
		update_bits_ &= ~ColorBit;
	}
}

void TriMesh::set_cv_mode(const std::string& _s)
{
	cv_mode_ = _s;
	update_later(ColorBit);
}

void TriMesh::set_cv_mode(VPropHandleFloat _p)
{
	prop_cv_handle_ = _p;
	set_cv_mode(prop_cv_mode_);
}
